from datamart_isi.profiler import Profiler
import pandas as pd
import typing
import warnings
import traceback
from datamart_isi.utilities.utils import Utils
from datamart_isi.joiners.joiner_base import JoinerPrepare, JoinerType
from datamart_isi.joiners.join_result import JoinResult
from SPARQLWrapper import SPARQLWrapper, JSON, POST, URLENCODED
from itertools import chain


class Augment(object):

    def __init__(self, endpoint: str) -> None:
        """Init method of QuerySystem, set up connection to elastic search.

        Args:
            endpoint: query endpoint address

        Returns:

        """

        self.qm = SPARQLWrapper(endpoint)
        self.qm.setReturnFormat(JSON)
        self.qm.setMethod(POST)
        self.qm.setRequestMethod(URLENCODED)

        self.joiners = dict()
        self.profiler = Profiler()

    def query_by_sparql(self, query: dict, dataset: pd.DataFrame=None, **kwargs) -> typing.Optional[typing.List[dict]]:
        """
        Args:
            query: a dictnary format query
            dataset: 
            **kwargs:

        Returns:

        """

        if query:
            query_body = self.parse_sparql_query(query, dataset)
            try:
                self.qm.setQuery(query_body)    
                results = self.qm.query().convert()['results']['bindings']
            except:
                print("[ERROR] Query failed!")
                traceback.print_exc()
                return []
            return results
        else:
            print("\n\n[ERROR] No query given, query failed!\n\n")
            return []

    def parse_sparql_query(self, json_query, dataset) -> str:
        """
        parse the json query to a spaqrl query format
        :param json_query:
        :param dataset:
        :return: a string indicate the sparql query
        """
        # example of query variables: Chaves Los Angeles Sacramento
        query_variables = json_query['variables']
        query_part = " ".join(query_variables.values())
        
        spaqrl_query = '''
        prefix ps: <http://www.wikidata.org/prop/statement/> 
        prefix pq: <http://www.wikidata.org/prop/qualifier/> 
        prefix p: <http://www.wikidata.org/prop/>

        SELECT ?dataset ?datasetLabel ?variableName ?variable ?score ?rank ?url ?file_type ?title ?keywords ?extra_information
        WHERE 
        {
          ?dataset p:C2005 ?variable;
                 rdfs:label ?datasetLabel.
          ?variable ps:C2005 ?variableName ;
                    pq:C2006 [
                        bds:search """''' + query_part + '''""" ;
                        bds:relevance ?score ;
                        bds:rank ?rank
                    ].
           ?dataset p:P2699/ps:P2699 ?url.
           ?dataset p:P2701/ps:P2701 ?file_type.
           ?dataset p:P1476/ps:P1476 ?title.
           ?dataset p:C2004/ps:C2004 ?keywords.
           ?dataset p:C2010/ps:C2010 ?extra_information.
          
        }
        ORDER BY DESC(?rank)
        LIMIT 30'''

        return spaqrl_query


    def query(self,
              col: pd.Series = None,
              minimum_should_match_ratio_for_col: float = None,
              query_string: str = None,
              temporal_coverage_start: str = None,
              temporal_coverage_end: str = None,
              global_datamart_id: int = None,
              variable_datamart_id: int = None,
              key_value_pairs: typing.List[tuple] = None,
              **kwargs
              ) -> typing.Optional[typing.List[dict]]:

        """Query metadata by a pandas Dataframe column

        Args:
            col: pandas Dataframe column.
            minimum_should_match_ratio_for_col: An float ranges from 0 to 1
                indicating the ratio of unique value of the column to be matched
            query_string: string to query any field in metadata
            temporal_coverage_start: start of a temporal coverage
            temporal_coverage_end: end of a temporal coverage
            global_datamart_id: match a global metadata id
            variable_datamart_id: match a variable metadata id
            key_value_pairs: match key value pairs

        Returns:
            matching docs of metadata
        """

        queries = list()

        if query_string:
            queries.append(
                self.qm.match_any(query_string=query_string)
            )

        if temporal_coverage_start or temporal_coverage_end:
            queries.append(
                self.qm.match_temporal_coverage(start=temporal_coverage_start, end=temporal_coverage_end)
            )

        if global_datamart_id:
            queries.append(
                self.qm.match_global_datamart_id(datamart_id=global_datamart_id)
            )

        if variable_datamart_id:
            queries.append(
                self.qm.match_variable_datamart_id(datamart_id=variable_datamart_id)
            )

        if key_value_pairs:
            queries.append(
                self.qm.match_key_value_pairs(key_value_pairs=key_value_pairs)
            )

        if col is not None:
            queries.append(
                self.qm.match_some_terms_from_variables_array(terms=col.unique().tolist(),
                                                              minimum_should_match=minimum_should_match_ratio_for_col)
            )

        if not queries:
            return self._query_all()

        return self.qm.search(body=self.qm.form_conjunction_query(queries), **kwargs)

    def join(self,
             left_df: pd.DataFrame,
             right_df: pd.DataFrame,
             left_columns: typing.List[typing.List[int]],
             right_columns: typing.List[typing.List[int]],
             left_metadata: dict = None,
             right_metadata: dict = None,
             joiner: JoinerType = JoinerType.DEFAULT
             ) -> JoinResult:

        """Join two dataframes based on different joiner.

          Args:
              left_df: pandas Dataframe
              right_df: pandas Dataframe
              left_metadata: metadata of left dataframe
              right_metadata: metadata of right dataframe
              left_columns: list of integers from left df for join
              right_columns: list of integers from right df for join
              joiner: string of joiner, default to be "default"

          Returns:
               JoinResult
          """

        if joiner not in self.joiners:
            self.joiners[joiner] = JoinerPrepare.prepare_joiner(joiner=joiner)

        if not self.joiners[joiner]:
            warnings.warn("No suitable joiner, return original dataframe")
            return JoinResult(left_df, [])

        print(" - start profiling")
        if not (left_metadata and left_metadata.get("variables")):
            # Left df is the user provided one.
            # We will generate metadata just based on the data itself, profiling and so on
            left_metadata = Utils.generate_metadata_from_dataframe(data=left_df, original_meta=left_metadata)

        if not right_metadata:
            right_metadata = Utils.generate_metadata_from_dataframe(data=right_df)

        # Only profile the joining columns, otherwise it will be too slow:
        left_metadata = Utils.calculate_dsbox_features(data=left_df, metadata=left_metadata,
                                                       selected_columns=set(chain.from_iterable(left_columns)))

        right_metadata = Utils.calculate_dsbox_features(data=right_df, metadata=right_metadata,
                                                        selected_columns=set(chain.from_iterable(right_columns)))

        # update with implicit_variable on the user supplied dataset
        if left_metadata.get('implicit_variables'):
            Utils.append_columns_for_implicit_variables_and_add_meta(left_metadata, left_df)

        print(" - start joining tables")
        res = self.joiners[joiner].join(left_df=left_df,
                                         right_df=right_df,
                                         left_columns=left_columns,
                                         right_columns=right_columns,
                                         left_metadata=left_metadata,
                                         right_metadata=right_metadata,
                                         )

        return res



import pandas as pd
import urllib.request
import typing
from datamart_isi.materializers.parsers.parse_result import ParseResult


class ParserBase:
    """
    Parser for GeneralMaterializer. the parsers will take the url for an HTML page, a csv file, a json file,\
    or an Excel file etc, as input, and parse them into pandas DataFrame(s) and return
    """
    def parse(self, url: str) -> typing.List[ParseResult]:
        return [ParseResult(df=res, index=idx) for idx, res in enumerate(self.get_all(url))]

    def get_all(self, url: str) -> typing.List[pd.DataFrame]:
        return []

    def get(self, url: str, index: int=0) -> typing.Optional[pd.DataFrame]:
        all_results = self.get_all(url)
        if all_results and index < len(all_results):
            return all_results[index]

    @staticmethod
    def load_content(url: str) -> bytes:
        return urllib.request.urlopen(url).read()





from datamart_isi.joiners.join_feature.feature_factory import *
import typing
import rltk
from rltk.io.reader.dataframe_reader import DataFrameReader

"""
@rltk.set_id("dataframe_default_index", function_=lambda s: str(s))
class LeftDynamicRecord(rltk.AutoGeneratedRecord):
    @property
    def id(self):
        return self.raw_object['id']

    @property
    def join_column(self):
        return self.raw_object['join_column']
"""



@rltk.set_id("dataframe_default_index", function_=lambda s: str(s))
class LeftDynamicRecord(rltk.AutoGeneratedRecord):
    pass


@rltk.set_id("dataframe_default_index", function_=lambda s: str(s))
class RightDynamicRecord(rltk.AutoGeneratedRecord):
    pass


class FeaturePairs:
    def __init__(self,
                 left_df: pd.DataFrame,
                 right_df: pd.DataFrame,
                 left_columns: typing.List[typing.List[int]],
                 right_columns: typing.List[typing.List[int]],
                 left_metadata: dict,
                 right_metadata: dict,
    ):
        l1 = len(left_columns)
        l2 = len(right_columns)
        if not (l1 == l2 and l1 and l2):
            # TODO: throw error or warning
            return

        self._length = l1

        self._left_df = left_df
        self._right_df = right_df
        self._left_columns = left_columns
        self._right_columns = right_columns
        self._left_metadata = left_metadata
        self._right_metadata = right_metadata

        self._left_rltk_dataset = self._init_rltk_dataset(left_df, LeftDynamicRecord)
        self._right_rltk_dataset = self._init_rltk_dataset(right_df, RightDynamicRecord)

        self._pairs = self._init_pairs()

    @property
    def left_rltk_dataset(self):
        return self._left_rltk_dataset

    @property
    def right_rltk_dataset(self):
        return self._right_rltk_dataset

    @property
    def pairs(self):
        return self._pairs

    def get_rltk_block(self) -> typing.Optional[rltk.Block]:
        prime_key_l = []
        prime_key_r = []
        str_key_l = []
        str_key_r = []
        for f1, f2 in self.pairs:
            if f1.data_type == DataType.STRING:
                # 2019.4.10: TOKEN_CATEGORICAL should also considered here
                if f1.distribute_type == DistributeType.CATEGORICAL or f1.distribute_type == DistributeType.TOKEN_CATEGORICAL:
                    prime_key_l.append(f1.name)
                    prime_key_r.append(f2.name)
                elif f1.distribute_type == DistributeType.NON_CATEGORICAL:
                    str_key_l.append(f1.name)
                    str_key_r.append(f2.name)

        if prime_key_l and prime_key_r:
            try:
                bg = rltk.HashBlockGenerator()
                block = bg.generate(
                    bg.block(self.left_rltk_dataset, function_=lambda r: ''.join([str(getattr(r, pk)).lower()
                                                                                  for pk in prime_key_l])),
                    bg.block(self.right_rltk_dataset, function_=lambda r: ''.join([str(getattr(r, pk)).lower()
                                                                                   for pk in prime_key_r])))
                return block
            except Exception as e:
                print(' - BLOCKING EXCEPTION: %s' % str(e))
                raise ValueError("failed to get blocking!")

        # if the datasets are too large, use each key's first char as blocking key
        if str_key_l and str_key_r and len(self._left_df) * len(self._right_df) > 10000:
            try:
                bg = rltk.HashBlockGenerator()
                block = bg.generate(
                    # original: str(getattr(r, pk))[0]
                    bg.block(self.left_rltk_dataset, function_=lambda r: ''.join([str(getattr(r, pk)).lower()
                                                                                  for pk in str_key_l])),
                    bg.block(self.right_rltk_dataset, function_=lambda r: ''.join([str(getattr(r, pk)).lower()
                                                                                   for pk in str_key_r]))
                    )
                return block
            except Exception as e:
                print(' - BLOCKING EXCEPTION: %s' % str(e))

    def __len__(self):
        return self._length

    def _init_pairs(self):
        return [(FeatureFactory.create(self._left_df, self._left_columns[i], self._left_metadata),
                 FeatureFactory.create(self._right_df, self._right_columns[i], self._right_metadata))
                for i in range(self._length)]

    @staticmethod
    def _init_rltk_dataset(df, record_class):
        rltk_dataset = rltk.Dataset(reader=DataFrameReader(df, True), record_class=record_class)
        return rltk_dataset




from abc import ABC, abstractmethod
from pandas import DataFrame
import typing


class MaterializerBase(ABC):
    """Abstract class of Materializer, should be extended for every Materializer dealing with different data source.

    """

    @abstractmethod
    def __init__(self, **kwargs):
        pass

    @abstractmethod
    def get(self,
            metadata: dict = None,
            constrains: dict = None
            ) -> typing.Optional[DataFrame]:
        """API for Materialize, every Materializer should implement `get` method, returns a pandas dataframe.

        """
        pass



from SPARQLWrapper import SPARQLWrapper, JSON
from argparse import ArgumentParser
import re
import json
import os
import typing

DEFAULT_SAVE_LOCATION = os.path.join(os.getcwd(), "prop_idents_v3.json")


def get_properties(pnode_id):
    """
    Function used to get the corresponding properties of given p node id
    :param pnode_id: a string to indicate a p node
    :return:
    """
    sparql = SPARQLWrapper("https://query.wikidata.org/sparql")
    cur_json = {}
    #sparql = SPARQLWrapper("http://sitaware.isi.edu:8080/bigdata/namespace/wdq/sparql")
    #sparql.setTimeout(1000)
    sparql.setQuery("""
        SELECT ?x ?v WHERE { ?x wdt:""" + pnode_id + """ ?v. }
    """)
    sparql.setReturnFormat(JSON)
    try:
        results = sparql.query().convert()
        for result in results["results"]["bindings"]:
            id = result["v"]["value"]
            Q = re.search(r'Q[0-9]+', result["x"]["value"])
            if Q:
                cur_json[id] = Q.group(0)
    except:
        print("failed to find properties on P node" + pnode_id)
    return cur_json


def get_identifiers() -> typing.List[str]:
    """
    The function used to find all P nodes
    :return:
    A list which contains all nodes
    """
    P_nodes = []
    sparql = SPARQLWrapper("https://query.wikidata.org/sparql")
    #sparql = SPARQLWrapper("http://sitaware.isi.edu:8080/bigdata/namespace/wdq/sparql")
    #sparql.setTimeout(1000)
    sparql.setQuery("""select ?p where {
        ?p wikibase:propertyType wikibase:ExternalId .
        }
        """)
    sparql.setReturnFormat(JSON)
    results = sparql.query().convert()
    res = results["results"]["bindings"]
    for r in res:
        try:
            P_node = re.search(r'[A-Za-z][0-9]+', r["p"]["value"]).group()
            P_nodes.append(P_node)
            # label = r["P_IDLabel"]["value"]
            # id_label[id]=label
        except:
            print("failed to find P nodes on " + str(r))
    return P_nodes


def save_prop_idents(save_location: str, test_mode: bool=False) -> None:
    """
    main function used to generate the property idents json structure
    :param save_location: path to the location
    :param test_mode: control whether to test on this function or run all
    :return:
    """
    print("The file will be stored at " + save_location)
    id_label = {}  # get_properties("P212")
    P_nodes = get_identifiers()
    prop_idents = {}

    if test_mode:
        print("In test mode, only run first 10 P-nodes")
        P_nodes = P_nodes[:10]

    len_p_nodes = len(P_nodes)

    for idx, P_node in enumerate(P_nodes):
        print("now processing " + P_node + " --> " + str(idx) + "/" + str(len_p_nodes))
        prop_idents[P_node] = get_properties(P_node)
    js = json.dumps(prop_idents)
    f = open("prop_idents_v3.json", "w+")
    f.write(js)
    f.close()


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument("-l", "--loc", action="store", type=str, dest="save_location", help="the location to save the file, including the file name")
    parser.add_argument("--test", action="store_const", const=True, dest="test_mode", help="whether to enter the test mode (only do process on first 10 p nodes) or not")
    args, _ = parser.parse_known_args()

    if args.save_location:
        save_location = args.save_location
    else:
        save_location = DEFAULT_SAVE_LOCATION

    save_prop_idents(save_location=save_location, test_mode=args.test_mode)




from datamart_isi.joiners.rltk_joiner import RLTKJoiner


class ExactMatchJoiner(RLTKJoiner):
    # TODO: work around for demo, should implement a better one

    def __init__(self):
        super().__init__()
        self.exact_match = True




import typing
import pandas as pd


class JoinResult(object):
    def __init__(self, df: pd.DataFrame or None, matched_rows: typing.List[int]=None):
        self._df = df
        self._matched_rows = matched_rows

    @property
    def df(self):
        return self._df

    @property
    def matched_rows(self):
        return self._matched_rows

    @property
    def cover_ratio(self):
        if self.matched_rows:
            cnt = len([_ for _ in self.matched_rows if _])
            if len(self.matched_rows):
                return cnt/len(self.matched_rows)


import pandas as pd
import typing
from datamart_isi.joiners.joiner_base import JoinerBase
from datamart_isi.joiners.join_feature.feature_pairs import FeaturePairs
from datamart_isi.joiners.join_result import JoinResult
import rltk

"""
TODO
Implement RLTK joiner
"""


class RLTKJoiner_new(JoinerBase):
    """
        This new joiner just simply join all found pairs in rltk's block
    """
    def __init__(self):
        # if set to be true, we can conduct exact match with rltk joiner
        self.exact_match = False 
        self.join_target_column_names = ""

    def set_join_target_column_names(self, column_name):
        self.join_target_column_names = column_name

    def find_pair(self,
                  left_df: pd.DataFrame,
                  right_df: pd.DataFrame) -> typing.Tuple[pd.DataFrame, typing.List[tuple]]:
        class left(rltk.AutoGeneratedRecord):
            pass

        class right(rltk.AutoGeneratedRecord):
            pass

        left_df['id']=left_df.index.astype(str)
        right_df['id']=right_df.index.astype(str)
        if 'Unnamed: 0' in right_df.columns:
            right_df = right_df.drop(columns=['Unnamed: 0'])
        ds1 = rltk.Dataset(rltk.DataFrameReader(left_df), record_class=left)
        ds2 = rltk.Dataset(rltk.DataFrameReader(right_df), record_class=right)

        bg = rltk.HashBlockGenerator()
        block = bg.generate(bg.block(ds1, property_=self.join_target_column_names[0]),
                            bg.block(ds2, property_=self.join_target_column_names[1]))

        pairs = rltk.get_record_pairs(ds1, ds2, block=block)

        pairs_list = []
        pairs_column = [[] for _ in range(right_df.shape[0])]
        for r1, r2 in pairs:
            pairs_column[int(r2.id)].append(int(r1.id))
            pairs_list.append((r1.id, r2.id))

        right_df["joining_pairs"] = pairs_column
        return right_df, pairs_list

    def join(self,
             left_df: pd.DataFrame,
             right_df: pd.DataFrame,
             left_columns: typing.List[typing.List[int]],
             right_columns: typing.List[typing.List[int]],
             left_metadata: dict,
             right_metadata: dict
             ) -> JoinResult:

        class left(rltk.AutoGeneratedRecord):
            pass

        class right(rltk.AutoGeneratedRecord):
            pass

        left_df['id']=left_df.index.astype(str)
        right_df['id']=right_df.index.astype(str)
        if 'Unnamed: 0' in right_df.columns:
            right_df = right_df.drop(columns=['Unnamed: 0'])
        ds1 = rltk.Dataset(rltk.DataFrameReader(left_df), record_class=left)
        ds2 = rltk.Dataset(rltk.DataFrameReader(right_df), record_class=right)

        bg = rltk.HashBlockGenerator()
        block = bg.generate(bg.block(ds1, property_=self.join_target_column_names[0]),
                            bg.block(ds2, property_=self.join_target_column_names[1]))
        left_df = left_df.set_index('id')
        right_df = right_df.set_index('id')

        pairs = rltk.get_record_pairs(ds1, ds2, block=block)

        df_joined = pd.DataFrame()

        column_names_to_join = None
        for r1, r2 in pairs:
            left_res = left_df.loc[r1.id]
            right_res = right_df.loc[r2.id]
            if column_names_to_join is None:
                column_names_to_join = right_res.index.difference(left_res.index)
                matched_rows = right_res.index.intersection(left_res.index)
                columns_new = left_res.index.tolist()
                columns_new.extend(column_names_to_join.tolist())
            new = pd.concat([left_res, right_res[column_names_to_join]])
            df_joined = df_joined.append(new, ignore_index = True)
        # ensure that the original dataframe columns are at the first left part
        df_joined = df_joined[columns_new]

        return JoinResult(df_joined, matched_rows)


class RLTKJoiner(JoinerBase):

    def __init__(self):
        pass

    @staticmethod
    def find_pair(
             left_df: pd.DataFrame,
             right_df: pd.DataFrame,
             left_columns: typing.List[typing.List[int]],
             right_columns: typing.List[typing.List[int]],
             left_metadata: dict,
             right_metadata: dict
             ) -> typing.Tuple[pd.DataFrame, typing.List[tuple]]:

        fp = FeaturePairs(left_df, right_df, left_columns, right_columns, left_metadata, right_metadata)
        block = fp.get_rltk_block()
        record_pairs = rltk.get_record_pairs(fp.left_rltk_dataset, fp.right_rltk_dataset, block=block)
        pairs_list = []
        # for r1, r2 in record_pairs:
        #     pairs_list.append((r1.id, r2.id))

        pairs_column = [[] for _ in range(right_df.shape[0])]
        for r1, r2 in record_pairs:
            pairs_column[int(r2.id)].append(int(r1.id))
            pairs_list.append((r1.id, r2.id))

        right_df["joining_pairs"] = pairs_column

        return right_df, pairs_list

    def join(self,
             left_df: pd.DataFrame,
             right_df: pd.DataFrame,
             left_columns: typing.List[typing.List[int]],
             right_columns: typing.List[typing.List[int]],
             left_metadata: dict,
             right_metadata: dict
             ) -> JoinResult:
        # print(left_metadata)

        # step 1 : transform columns
        """
        1. merge columns if multi-multi relation, mark as "merged" so we use set-based functions only
        2. pull out the mapped columns and form to new datasets with same order to support
        """
        fp = FeaturePairs(left_df, right_df, left_columns, right_columns, left_metadata, right_metadata)
        block = fp.get_rltk_block()

        # test codes here for taking a look inside of block
        # for idx, b in enumerate(block.key_set_adapter):
        #     print(b)
        #     if idx == 10:
        #         break

        record_pairs = rltk.get_record_pairs(fp.left_rltk_dataset, fp.right_rltk_dataset, block=block)

        # similarity matrix, the shape is (left_df.shape[0] * right_df.shape[0])
        sim = [[0 for _ in range(len(right_df))] for _ in range(len(left_df))]
        # enumerate the record pairs and calculate the similarity
        for r1, r2 in record_pairs:
            similarities = []
            for f1, f2 in fp.pairs:
                v1 = f1.value_merge_func(r1)
                v2 = f2.value_merge_func(r2)
                if self.exact_match:
                    similarities.append(1 if v1 == v2 else 0)
                    continue
                # print(v1, v2, type(f1), type(f2))
                for similarity_func in f1.similarity_functions():
                    similarity = similarity_func(v1, v2)
                    similarities.append(similarity)
                    # print(f1.name, f2.name, v1, v2, similarity, similarity_func, type(f1))
                    # TODO: now only consider the first similarity function for now
                    break
                # print(v1, v2, similarities)
            sim[int(r1.id)][int(r2.id)] = sum(similarities)/len(similarities) if similarities else 0

        # choose the highest similarity row as the matched rows, the matched_rows have same length as left_df.shape[0]
        # TODO: why only choose the highest similarity? maybe considering add some other methods
        matched_rows = self.simple_best_match(sim)
        res = self.one_to_one_concat(matched_rows, left_df, right_df, right_columns)

        # step 2 : analyze target columns - get ranked similarity functions for each columns
        """
        see https://paper.dropbox.com/doc/ER-for-Datamart--ASlKtpR4ceGaj~6cN4Q7EWoSAQ-tRug6oRX6g5Ko5jzaeynT
        """

        # step 3 : check if 1-1, 1-n, n-1, or m-n relations,
        # based on the analyze in step 2 we can basically know if it is one-to-one relation

        return JoinResult(res, matched_rows)

    def one_to_one_concat(self, matched_rows, left_df, right_df, right_columns):
        right_remain = self.get_remain_list(right_df, right_columns)
        right_remain_headers = [right_df.columns.tolist()[j] for j in right_remain]
        series_arr = []
        for i in matched_rows:
            if isinstance(i, int):
                series_arr.append(right_df.iloc[i, right_remain])
            else:
                series_arr.append(pd.Series(index=right_remain_headers))
        to_join = pd.DataFrame(series_arr, index=range(len(matched_rows)))
        res = pd.concat([left_df, to_join], axis=1)
        return res

    def munkrus_match(self, sim: typing.List[typing.List[float]]):
        pass

    @staticmethod
    def simple_best_match(sim: typing.List[typing.List[float]], threshold=0.5):
        res = []
        for idx, v in enumerate(sim):
            max_val = threshold
            max_idx = None
            for idx_, v_ in enumerate(v):
                if v_ >= max_val:
                    max_idx = idx_
                    max_val = v_
            res.append(max_idx)
        return res

    @staticmethod
    def simple_best_matches(sim: typing.List[typing.List[float]], threshold=0.5):
        res = []
        for idx, v in enumerate(sim):
            cur = []
            for idx_, v_ in enumerate(v):
                if v_ >= threshold:
                    cur.append(idx_)
            res.append(cur)
        return res

    @staticmethod
    def get_remain_list(df: pd.DataFrame, columns_2d: typing.List[typing.List[int]]):
        all_columns = list(range(df.shape[1]))
        columns_1d = [item for sublist in columns_2d for item in sublist]
        remaining = [_ for _ in all_columns if _ not in columns_1d]
        return remaining






from SPARQLWrapper import SPARQLWrapper, JSON,XML
import re,json

#get all identifiers, i.e.[P1,P2,P3, ...]
def get_identifiers():
	global P_nodes
	#sparql = SPARQLWrapper("https://query.wikidata.org/sparql")
	sparql = SPARQLWrapper("http://sitaware.isi.edu:8080/bigdata/namespace/wdq/sparql")
	#sparql.setTimeout(1000)
	sparql.setQuery("""select ?p where {
		?p wikibase:propertyType wikibase:ExternalId .
		}
		""")
	sparql.setReturnFormat(JSON)
	results = sparql.query().convert()
	res = results["results"]["bindings"]
	for r in res:
		try:
			P_node = re.search(r'[A-Za-z][0-9]+',r["p"]["value"]).group()
			#update global varibale P_nodes
			P_nodes.append(P_node)
		except:
			print('get_identifiers Error()', r)

#get properties for a given P_node, return {'01101':'Q1234567'}
def get_properties(P_ID):
	cur_json = {}
	#sparql = SPARQLWrapper("https://query.wikidata.org/sparql")
	sparql = SPARQLWrapper("http://sitaware.isi.edu:8080/bigdata/namespace/wdq/sparql")
	#sparql.setTimeout(1000)
	sparql.setQuery("""
		SELECT ?x ?v WHERE { ?x wdt:""" + P_ID + """ ?v. }
	""")
	sparql.setReturnFormat(JSON)
	try:
		results = sparql.query().convert()
		for result in results["results"]["bindings"]:
			value = result["v"]["value"]
			Q = re.search(r'Q[0-9]+',result["x"]["value"])
			if Q:
				Q_node = Q.group(0)
				cur_json[value] = Q_node
				#if the key is a number and has leading zeros, add a new key in the dict
				if(value[0]=="0" and value.isdigit()):
					rm_leading0 = str(int(value))
					cur_json[rm_leading0] = Q_node
	except:
		print('get_properties() Error:',P_ID)
	return cur_json

#do traversal on prop_idents dict, build identifier_nodes_dict, i.e. {'1101': [P882/Q123456,PP881/Q123]}
def convert_dict(prop_idents):
    identifier_nodes_dict = {}
    count = 0
    keys = set()
    for P_number,p_entity in prop_idents.items():
        count += 1
        print(P_number,count)
        for identi,Q_number in p_entity.items():
            #if identi in identifier_nodes_dict.keys():
            if identi in keys:
                identifier_nodes_dict[identi].append(P_number + '/' + Q_number)
            else:
                keys.add(identi)
                identifier_nodes_dict[identi] = [P_number + '/' + Q_number]
    return identifier_nodes_dict

if __name__ == "__main__":
	P_nodes = []
	prop_idents ={}
	get_identifiers()
	for idx,P_node in enumerate(P_nodes):
		print(idx,P_node)
		prop_idents[P_node] = get_properties(P_node)

	identifier_nodes = convert_dict(prop_idents)
	js = json.dumps(identifier_nodes)
	# could be changed to any file path
	f = open("identifier_nodes.json","w+")
	f.write(js)
	f.close()



from rltk.similarity import *


def to_lower(x: list):
    return [_.lower() for _ in x]


def to_set(x: list, ignore_cases: bool=True):
    if ignore_cases:
        return set([_.lower() for _ in x])
    return set(x)


def to_str(x: list, joiner: str=' '):
    return joiner.join(x)


def jaccard_sim(left: list, right: list):
    return jaccard_index_similarity(to_set(left), to_set(right))


def hybrid_jaccard_sim(left: list, right: list):
    return hybrid_jaccard_similarity(to_set(left), set(right))


def levenshtein_sim(left: list, right: list):
    return levenshtein_similarity(to_str(left), to_str(right))


def jaro_winkler_sim(left: list, right: list):
    return jaro_winkler_similarity(to_str(left), to_str(right))


def ngram_sim(left: list, right: list):
    return ngram_similarity(to_str(left), to_str(right))


def cosine_sim(left: list, right: list):
    return cosine_similarity(left, right)


from datamart_isi.metadata.metadata_base import MetadataBase
from datamart_isi.utilities.utils import Utils
import typing


class VariableMetadata(MetadataBase):
    def __init__(self, description: dict, datamart_id: typing.Union[int, None] = None) -> None:
        """Init method of VariableMetadata.

        Args:
            description: description dict.
            datamart_id: unique datamart_id.

        Returns:

        """

        super().__init__()

        self._metadata["datamart_id"] = datamart_id

        if "name" in description:
            self._metadata["name"] = description["name"]

        if "description" in description:
            self._metadata["description"] = description["description"]

        self._metadata["semantic_type"] = description.get("semantic_type", [])

        if "named_entity" in description:
            self._metadata["named_entity"] = description["named_entity"]

        if "temporal_coverage" in description:
            self._metadata["temporal_coverage"] = description["temporal_coverage"]

        if self.temporal_coverage is not False:
            self.temporal_coverage = Utils.temporal_coverage_validate(self.temporal_coverage)

        if "spatial_coverage" in description:
            self._metadata["spatial_coverage"] = description["spatial_coverage"]

    @classmethod
    def construct_variable(cls, description, datamart_id=None) -> 'VariableMetadata':
        return cls(description, datamart_id)

    @property
    def datamart_id(self):
        return self._metadata["datamart_id"]

    @datamart_id.setter
    def datamart_id(self, value):
        self._metadata["datamart_id"] = value

    @property
    def name(self):
        return self._metadata.get("name", False)

    @name.setter
    def name(self, value):
        self._metadata["name"] = value

    @property
    def description(self):
        return self._metadata.get("description", False)

    @description.setter
    def description(self, value):
        self._metadata["description"] = value

    @property
    def semantic_type(self):
        return self._metadata["semantic_type"]

    @semantic_type.setter
    def semantic_type(self, value):
        self._metadata["semantic_type"] = value

    @property
    def named_entity(self):
        return self._metadata.get("named_entity", False)

    @named_entity.setter
    def named_entity(self, value):
        self._metadata["named_entity"] = value

    @property
    def temporal_coverage(self):
        return self._metadata.get("temporal_coverage", False)

    @temporal_coverage.setter
    def temporal_coverage(self, value):
        self._metadata["temporal_coverage"] = value

    @property
    def spatial_coverage(self):
        return self._metadata.get("spatial_coverage", False)

    @spatial_coverage.setter
    def spatial_coverage(self, value):
        self._metadata["spatial_coverage"] = value



from abc import ABC, abstractmethod
from typing import Dict


class MetadataBase(ABC):
    """Abstract class of Metadata, should be extended for global and variable metadata.

    """

    @abstractmethod
    def __init__(self) -> None:
        self._metadata = dict()

    @property
    def value(self) -> Dict:
        """Return actual json representation of metadata that will be load to ES

        """
        return self._metadata




import typing
import pandas as pd
import copy
import random
import frozendict
import collections
import typing
import traceback
import logging
import datetime
import d3m.metadata.base as metadata_base
import json

from ast import literal_eval
from SPARQLWrapper import SPARQLWrapper, JSON, POST, URLENCODED
from d3m import container
from d3m import utils
from d3m.container import DataFrame as d3m_DataFrame
from d3m.container import Dataset as d3m_Dataset
from d3m.base import utils as d3m_utils
from d3m.metadata.base import DataMetadata, ALL_ELEMENTS

from datamart_isi.augment import Augment
from datamart_isi.utilities.utils import Utils
from datamart_isi.joiners.rltk_joiner import RLTKJoiner
from datamart_isi.joiners.rltk_joiner import RLTKJoiner_new

# from datamart.joiners.join_result import JoinResult
# from datamart.joiners.joiner_base import JoinerType

# this is a private file
from datamart_isi import config


__all__ = ('DatamartQueryCursor', 'Datamart', 'DatasetColumn', 'DatamartSearchResult', 'AugmentSpec',
           'TabularJoinSpec', 'UnionSpec', 'TemporalGranularity', 'GeospatialGranularity', 'ColumnRelationship', 'DatamartQuery',
           'VariableConstraint', 'NamedEntityVariable', 'TemporalVariable', 'GeospatialVariable', 'TabularVariable')

Q_NODE_SEMANTIC_TYPE = "http://wikidata.org/qnode"
AUGMENTED_COLUMN_SEMANTIC_TYPE = "https://metadata.datadrivendiscovery.org/types/Datamart_augmented_column"
MAX_ENTITIES_LENGTH = 200
CONTAINER_SCHEMA_VERSION = 'https://metadata.datadrivendiscovery.org/schemas/v0/container.json'
P_NODE_IGNORE_LIST = {"P1549"}
SPECIAL_REQUEST_FOR_P_NODE = {"P1813": "FILTER(strlen(str(?P1813)) = 2)"}
AUGMENT_RESOURCE_ID = "learningData"
WIKIDATA_QUERY_SERVER = config.wikidata_server


class DatamartQueryCursor(object):
    """
    Cursor to iterate through Datamarts search results.
    """
    def __init__(self, augmenter, search_query, supplied_data, need_run_wikifier=None):
        self.augmenter = augmenter
        self.search_query = search_query
        self.current_searching_query_index = 0
        self.supplied_data = supplied_data
        self.remained_part = None
        if need_run_wikifier is None:
            self.need_run_wikifier = self._check_need_wikifier_or_not()
        else:
            self.need_run_wikifier = need_run_wikifier

    def get_next_page(self, *, limit: typing.Optional[int] = 20, timeout: int = None) \
            -> typing.Optional[typing.Sequence['DatamartSearchResult']]:
        """
        Return the next page of results. The call will block until the results are ready.

        Note that the results are not ordered; the first page of results can be returned first simply because it was
        found faster, but the next page might contain better results. The caller should make sure to check
        `DatamartSearchResult.score()`.

        Parameters
        ----------
        limit : int or None
            Maximum number of search results to return. None means no limit.
        timeout : int
            Maximum number of seconds before returning results. An empty list might be returned if it is reached.

        Returns
        -------
        Sequence[DatamartSearchResult] or None
            A list of `DatamartSearchResult's, or None if there are no more results.
        """

        # if need to run wikifier, run it before any search
        if self.current_searching_query_index == 0 and self.need_run_wikifier:
            self._run_wikifier()

        # if already remained enough part
        current_result = self.remained_part or []
        if len(current_result) > limit:
            self.remained_part = current_result[limit:]
            current_result = current_result[:limit]
            return current_result

        # start searching
        while self.current_searching_query_index < len(self.search_query) and len(current_result) < limit:
            if self.search_query[self.current_searching_query_index].search_type == "wikidata":
                # TODO: now wikifier can only automatically search for all possible columns and do exact match
                search_res = self._search_wikidata()
            elif self.search_query[self.current_searching_query_index].search_type == "general":
                search_res = self._search_datamart()
            else:
                raise ValueError("Unknown search query type for " +
                                 self.search_query[self.current_searching_query_index].search_type)

            self.current_searching_query_index += 1
            current_result.extend(search_res)

        if len(current_result) == 0:
            return None
        else:
            if len(current_result) > limit:
                self.remained_part = current_result[limit:]
                current_result = current_result[:limit]
            return current_result

    def _check_need_wikifier_or_not(self) -> bool:
        """
        inner function used to check whether there already exist Q nodes in the supplied ata
        If not, will return True to indicate need to run wikifier
        If already exist Q nodes, return False to indicate no need to run wikifier
        :return: a bool value to indicate whether need to run wikifier or not
        """
        if type(self.supplied_data) is d3m_Dataset:
            res_id, supplied_dataframe = d3m_utils.get_tabular_resource(dataset=self.supplied_data, resource_id=None)
            selector_base_type = "ds"
        else:
            supplied_dataframe = self.supplied_data
            selector_base_type = "df"

        for i in range(supplied_dataframe.shape[1]):
            if selector_base_type == "ds":
                metadata_selector = (res_id, metadata_base.ALL_ELEMENTS, i)
            else:
                metadata_selector = (metadata_base.ALL_ELEMENTS, i)
            each_metadata = self.supplied_data.metadata.query(metadata_selector)
            if Q_NODE_SEMANTIC_TYPE in each_metadata['semantic_types']:
                print("[INFO] Q nodes columns found in input data, will not run wikifier.")
                return False
        print("[INFO] No Q nodes columns found in input data, will run wikifier.")
        return True

    def _run_wikifier(self):
        """
        function used to run wikifier primitive
        :return:
        """
        from dsbox.datapreprocessing.cleaner.wikifier import WikifierHyperparams, Wikifier
        wikifier_hyperparams = WikifierHyperparams.defaults()
        wikifier_primitive = Wikifier(hyperparams=wikifier_hyperparams)
        self.supplied_data = wikifier_primitive.produce(inputs=self.supplied_data).value
        self.need_run_wikifier = False

    def _search_wikidata(self, query=None, supplied_data: typing.Union[d3m_DataFrame, d3m_Dataset]=None, timeout=None,
                         search_threshold=0.5) -> typing.List["DatamartSearchResult"]:
        """
        The search function used for wikidata search
        :param query: JSON object describing the query.
        :param supplied_data: the data you are trying to augment.
        :param timeout: allowed time spent on searching
        :param limit: the limitation on the return amount of DatamartSearchResult
        :param search_threshold: the minimum appeared times of the properties
        :return: list of search results of DatamartSearchResult
        """
        if supplied_data is None:
            supplied_data = self.supplied_data

        wikidata_results = []
        try:
            q_nodes_columns = []
            if type(supplied_data) is d3m_Dataset:
                res_id, supplied_dataframe = d3m_utils.get_tabular_resource(dataset=supplied_data, resource_id=None)
                selector_base_type = "ds"
            else:
                supplied_dataframe = supplied_data
                selector_base_type = "df"

            # check whether Qnode is given in the inputs, if given, use this to wikidata and search
            required_variables_names = None
            metadata_input = supplied_data.metadata

            if query is not None and 'required_variables' in query:
                required_variables_names = []
                for each in query['required_variables']:
                    required_variables_names.extend(each['names'])
            for i in range(supplied_dataframe.shape[1]):
                if selector_base_type == "ds":
                    metadata_selector = (res_id, metadata_base.ALL_ELEMENTS, i)
                else:
                    metadata_selector = (metadata_base.ALL_ELEMENTS, i)
                if Q_NODE_SEMANTIC_TYPE in metadata_input.query(metadata_selector)["semantic_types"]:
                    # if no required variables given, attach any Q nodes found
                    if required_variables_names is None:
                        q_nodes_columns.append(i)
                    # otherwise this column has to be inside required_variables
                    else:
                        if supplied_dataframe.columns[i] in required_variables_names:
                            q_nodes_columns.append(i)

            if len(q_nodes_columns) == 0:
                print("No wikidata Q nodes detected on corresponding required_variables! Will skip wikidata search part")
                return wikidata_results
            else:

                print("Wikidata Q nodes inputs detected! Will search with it.")
                print("Totally " + str(len(q_nodes_columns)) + " Q nodes columns detected!")

                # do a wikidata search for each Q nodes column
                for each_column in q_nodes_columns:
                    q_nodes_list = supplied_dataframe.iloc[:, each_column].tolist()
                    p_count = collections.defaultdict(int)
                    p_nodes_needed = []
                    # temporary block
                    """
                    http_address = 'http://minds03.isi.edu:4444/get_properties'
                    headers = {"Content-Type": "application/json"}
                    requests_data = str(q_nodes_list)
                    requests_data = requests_data.replace("'", '"')
                    r = requests.post(http_address, data=requests_data, headers=headers)
                    results = r.json()
                    for each_p_list in results.values():
                        for each_p in each_p_list:
                            p_count[each_p] += 1
                    """
                    # TODO: temporary change here, may change back in the future
                    # Q node format (wd:Q23)(wd: Q42)
                    q_node_query_part = ""
                    unique_qnodes = set(q_nodes_list)
                    for each in unique_qnodes:
                        if len(each) > 0:
                            q_node_query_part += "(wd:" + each + ")"
                    sparql_query = "select distinct ?item ?property where \n{\n  VALUES (?item) {" + q_node_query_part \
                                   + "  }\n  ?item ?property ?value .\n  ?wd_property wikibase:directClaim ?property ." \
                                   + "  values ( ?type ) \n  {\n    ( wikibase:Quantity )\n" \
                                   + "    ( wikibase:Time )\n    ( wikibase:Monolingualtext )\n  }" \
                                   + "  ?wd_property wikibase:propertyType ?type .\n}\norder by ?item ?property "

                    try:
                        sparql = SPARQLWrapper(WIKIDATA_QUERY_SERVER)
                        sparql.setQuery(sparql_query)
                        sparql.setReturnFormat(JSON)
                        sparql.setMethod(POST)
                        sparql.setRequestMethod(URLENCODED)
                        results = sparql.query().convert()['results']['bindings']
                    except:
                        print("Query failed!")
                        traceback.print_exc()
                        continue

                    for each in results:
                        p_count[each['property']['value'].split("/")[-1]] += 1

                    for key, val in p_count.items():
                        if float(val) / len(unique_qnodes) >= search_threshold:
                            p_nodes_needed.append(key)
                    wikidata_search_result = {"p_nodes_needed": p_nodes_needed,
                                              "target_q_node_column_name": supplied_dataframe.columns[each_column]}
                    wikidata_results.append(DatamartSearchResult(search_result=wikidata_search_result,
                                                                 supplied_data=supplied_data,
                                                                 query_json=query,
                                                                 search_type="wikidata")
                                            )
            return wikidata_results

        except:
            print("Searching with wiki data failed")
            traceback.print_exc()
        finally:
            return wikidata_results

    def _search_datamart(self):
        """
        function used for searching in datamart with triple store database
        :return:
        """
        search_result = []
        query = {"keywords": self.search_query[self.current_searching_query_index].keywords,
                 "variables": self.search_query[self.current_searching_query_index].variables
                 }
        query_results = self.augmenter.query_by_sparql(query=query, dataset=self.supplied_data)
        for each in query_results:
            temp = DatamartSearchResult(search_result=each, supplied_data=self.supplied_data, query_json=query, search_type="general")
            search_result.append(temp)

        return search_result


class Datamart(object):
    """
    ISI implement of datamart
    """

    def __init__(self, connection_url: str) -> None:
        self.connection_url = connection_url
        self._logger = logging.getLogger(__name__)
        # query_server = "http://dsbox02.isi.edu:9001/blazegraph/namespace/datamart3/sparql"
        self.augmenter = Augment(endpoint=self.connection_url)
        self.supplied_dataframe = None

    def set_test_mode(self) -> None:
        query_server = config.wikidata_server_test
        self.augmenter = Augment(endpoint=query_server)

    def search(self, query: 'DatamartQuery') -> DatamartQueryCursor:
        """This entry point supports search using a query specification.

        The query specification supports querying datasets by keywords, named entities, temporal ranges, and geospatial ranges.

        Datamart implementations should return a DatamartQueryCursor immediately.

        Parameters
        ----------
        query : DatamartQuery
            Query specification.

        Returns
        -------
        DatamartQueryCursor
            A cursor pointing to search results.
        """
        print("Not implemented yet")
        pass

    def search_with_data(self, query: 'DatamartQuery', supplied_data: container.Dataset, skip_wikidata=False) -> DatamartQueryCursor:
        """
        Search using on a query and a supplied dataset.

        This method is a "smart" search, which leaves the Datamart to determine how to evaluate the relevance of search
        result with regard to the supplied data. For example, a Datamart may try to identify named entities and date
        ranges in the supplied data and search for companion datasets which overlap.

        To manually specify query constraints using columns of the supplied data, use the `search_with_data_columns()`
        method and `TabularVariable` constraints.

        Datamart implementations should return a DatamartQueryCursor immediately.

        Parameters
        ------_---
        query : DatamartQuery
            Query specification
        supplied_data : container.Dataset
            The data you are trying to augment.

        Returns
        -------
        DatamartQueryCursor
            A cursor pointing to search results containing possible companion datasets for the supplied data.
        """

        # first take a search on wikidata
        # add wikidata searching query at first position
        if skip_wikidata:
            search_queries = []
        else:
            search_queries = [DatamartQuery(search_type="wikidata")]

        if type(supplied_data) is d3m_Dataset:
            res_id, self.supplied_dataframe = d3m_utils.get_tabular_resource(dataset=supplied_data, resource_id=None)
        else:
            self.supplied_dataframe = supplied_data

        if query is None:
            # if not query given, try to find the Text columns from given dataframe and use it to find some candidates
            can_query_columns = []
            for each in range(len(self.supplied_dataframe.columns)):
                if type(supplied_data) is d3m_Dataset:
                    selector = (res_id, ALL_ELEMENTS, each)
                else:
                    selector = (ALL_ELEMENTS, each)
                each_column_meta = supplied_data.metadata.query(selector)
                if 'http://schema.org/Text' in each_column_meta["semantic_types"]:
                    # or "https://metadata.datadrivendiscovery.org/types/CategoricalData" in each_column_meta["semantic_types"]:
                    can_query_columns.append(each)

            if len(can_query_columns) == 0:
                self._logger.warning("No columns can be augment with datamart!")
            
            for each_column in can_query_columns:
                tabular_variable = TabularVariable(columns=[each_column], relationship=None)
                each_search_query = self._generate_datamart_query_from_data(query=None, supplied_data=supplied_data, data_constraints=tabular_variable)
                search_queries.append(each_search_query)

            return DatamartQueryCursor(augmenter=self.augmenter, search_query=search_queries, supplied_data=supplied_data)

    def search_with_data_columns(self, query: 'DatamartQuery', supplied_data: container.Dataset,
                                 data_constraints: typing.List['TabularVariable']) -> DatamartQueryCursor:
        """
        Search using a query which can include constraints on supplied data columns (TabularVariable).

        This search is similar to the "smart" search provided by `search_with_data()`, but caller must manually specify
        constraints using columns from the supplied data; Datamart will not automatically analyze it to determine
        relevance or joinability.

        Use of the query spec enables callers to compose their own "smart search" implementations.

        Datamart implementations should return a DatamartQueryCursor immediately.

        Parameters
        ------_---
        query : DatamartQuery
            Query specification
        supplied_data : container.Dataset
            The data you are trying to augment.
        data_constraints : list
            List of `TabularVariable` constraints referencing the supplied data.

        Returns
        -------
        DatamartQueryCursor
            A cursor pointing to search results containing possible companion datasets for the supplied data.
        """

        # put the enetities of all given columns from "data_constraints" into the query's variable part and run the query

        search_query = self._generate_datamart_query_from_data(query=None, supplied_data=supplied_data, data_constraints=data_constraints)
        return DatamartQueryCursor(augmenter=self.augmenter, search_query=[search_query], supplied_data=supplied_data)

    def _generate_datamart_query_from_data(self, query: 'DatamartQuery', supplied_data: container.Dataset,
                                 data_constraints: typing.List['TabularVariable']):
        all_query_variables = {}
        for each_column in data_constraints.columns:
            column_values = self.supplied_dataframe.iloc[:, each_column].astype(str)
            query_column_entities = list(set(column_values.tolist()))

            if len(query_column_entities) > MAX_ENTITIES_LENGTH:
                query_column_entities = random.sample(query_column_entities, MAX_ENTITIES_LENGTH)

            query_column_entities = " ".join(query_column_entities)
            all_query_variables[self.supplied_dataframe.columns[each_column]] = query_column_entities

        search_query = DatamartQuery(variables=all_query_variables)

        return search_query


class DatasetColumn:
    """
    Specify a column of a dataframe in a D3MDataset
    """

    def __init__(self, resource_id: str, column_index: int) -> None:
        self.resource_id = resource_id
        self.column_index = column_index


class DatamartSearchResult:
    """
    This class represents the search results of a datamart search.
    Different datamarts will provide different implementations of this class.

    Attributes
    ----------
    join_hints: typing.List[D3MAugmentSpec]
        Hints for joining supplied data with datamart data

    """
    def __init__(self, search_result, supplied_data, query_json, search_type):
        self._logger = logging.getLogger(__name__)
        self.search_result = search_result
        if "score" in self.search_result:
            self._score = float(self.search_result["score"]["value"])
        else:
            self._score = 1
        self.supplied_data = supplied_data
        if type(supplied_data) is d3m_Dataset:
            self.res_id, self.supplied_dataframe = d3m_utils.get_tabular_resource(dataset=supplied_data, resource_id=None)
            self.selector_base_type = "ds"
        elif type(supplied_data) is d3m_DataFrame:
            self.supplied_dataframe = supplied_data
            self.selector_base_type = "df"
        else:
            self.supplied_dataframe = None
        self.connection_url = WIKIDATA_QUERY_SERVER
        self.query_json = query_json
        self.search_type = search_type
        self.pairs = None
        self._res_id = None  # only used for input is Dataset
        self.join_pairs = None
        self.right_df = None
        self.d3m_metadata = self._get_d3m_metadata()

    def _get_d3m_metadata(self) -> DataMetadata:
        """
        function used to generate the d3m format metadata
        """
        if self.search_type == "wikidata":
            return self._get_d3m_metadata_for_wikidata()
        elif self.search_type == "general":
            return self._get_d3m_metadata_for_general()
        elif self.search_type == "wikifier":
            print("[WARN] No metadata can provide for wikifier augment")
            return DataMetadata()
        else:
            print("[ERROR] Unknown search type!")
            return DataMetadata()

    def _get_d3m_metadata_for_wikidata(self):
        """
        function used to generate the d3m format metadata - specified for wikidata search result
        because search results don't have value type of each P node, we have to query one sample to find
        """
        return_metadata = DataMetadata()
        if self.supplied_dataframe is not None:
            data_length = self.supplied_dataframe.shape[0]
        else:
            print("[WARNING] Can't calculate the row length for wikidata search results without supplied data")
            data_length = 1
        metadata_all = {"structural_type": d3m_DataFrame,
                        "semantic_types": ["https://metadata.datadrivendiscovery.org/types/Table"],
                        "dimension": {
                            "name": "rows",
                            "semantic_types": ["https://metadata.datadrivendiscovery.org/types/TabularRow"],
                            "length": data_length,
                            },
                        "schema": "https://metadata.datadrivendiscovery.org/schemas/v0/container.json"
                        }
        return_metadata = return_metadata.update(selector=(), metadata=metadata_all)
        metadata_all_elements = {
            "dimension": {
                "name": "columns",
                "semantic_types": ["https://metadata.datadrivendiscovery.org/types/TabularColumn"],
                "length": len(self.search_result['p_nodes_needed']),
            }
        }
        return_metadata = return_metadata.update(selector=(ALL_ELEMENTS,), metadata=metadata_all_elements)

        for i, each in enumerate(self.search_result['p_nodes_needed']):
            target_q_node_column_name = self.search_result['target_q_node_column_name']
            try:

                q_node_column_number = self.supplied_dataframe.columns.tolist().index(target_q_node_column_name)
                sample_row_number = 0
                q_node_sample = self.supplied_dataframe.iloc[sample_row_number, q_node_column_number]
                semantic_types = self._get_wikidata_column_semantic_types(q_node_sample, each)
                # if we failed with first test, repeat until we get success one
                while not semantic_types[0]:
                    sample_row_number += 1
                    q_node_sample = self.supplied_dataframe.iloc[sample_row_number, q_node_column_number]
                    semantic_types = self._get_wikidata_column_semantic_types(q_node_sample, each)
            except:
                semantic_types = (
                    'https://metadata.datadrivendiscovery.org/types/Attribute',
                    AUGMENTED_COLUMN_SEMANTIC_TYPE
                )
            each_metadata = {
                "name": self._get_node_name(self.search_result['p_nodes_needed'][i]) + "_for_" + target_q_node_column_name,
                "structural_type": str,
                "semantic_types": semantic_types,
            }
            return_metadata = return_metadata.update(selector=(ALL_ELEMENTS, i), metadata=each_metadata)

            each_metadata = {
                "name": "q_node",
                "structural_type": str,
                "semantic_types": (
                    'http://schema.org/Text',
                    'https://metadata.datadrivendiscovery.org/types/Attribute',
                    Q_NODE_SEMANTIC_TYPE,
                    AUGMENTED_COLUMN_SEMANTIC_TYPE
                ),
            }
            return_metadata = return_metadata.update(selector=(ALL_ELEMENTS, i+1), metadata=each_metadata)

            each_metadata = {
                "name": "joining_pairs",
                "structural_type": list,
                "semantic_types": (
                    'https://metadata.datadrivendiscovery.org/types/Attribute',
                    AUGMENTED_COLUMN_SEMANTIC_TYPE
                ),
            }
            return_metadata = return_metadata.update(selector=(ALL_ELEMENTS, i+2), metadata=each_metadata)
        return return_metadata

    def _get_wikidata_column_semantic_types(self, q_node_sample, p_node_target):
        """
        Inner function used to get the semantic types for given wikidata column
        :return: a tuple, tuple[0] indicate success get semantic type or not, tuple[1] indicate the found semantic types
        """
        q_nodes_query = '(wd:' + q_node_sample + ') \n'
        p_nodes_query_part = ' ?' + p_node_target + '\n'
        p_nodes_optional_part = "  OPTIONAL { ?q wdt:" + p_node_target + " ?" + p_node_target + "}\n"
        sparql_query = "SELECT DISTINCT ?q " + p_nodes_query_part + \
                       "WHERE \n{\n  VALUES (?q) { \n " + q_nodes_query + "}\n" + \
                       p_nodes_optional_part + "}\n"
        try:
            sparql = SPARQLWrapper(WIKIDATA_QUERY_SERVER)
            sparql.setQuery(sparql_query)
            sparql.setReturnFormat(JSON)
            sparql.setMethod(POST)
            sparql.setRequestMethod(URLENCODED)
            p_val = sparql.query().convert()['results']['bindings'][0][p_node_target]
            if "datatype" in p_val.keys():
                semantic_types = (
                    self._get_semantic_type(p_val["datatype"]),
                    'https://metadata.datadrivendiscovery.org/types/Attribute', AUGMENTED_COLUMN_SEMANTIC_TYPE)
            else:
                semantic_types = (
                    "http://schema.org/Text", 'https://metadata.datadrivendiscovery.org/types/Attribute',
                    AUGMENTED_COLUMN_SEMANTIC_TYPE)
            return True, semantic_types
        except:
            return False, None

    def _get_d3m_metadata_for_general(self):
        """
        function used to generate the d3m format metadata - specified for general search result
        """
        return_metadata = DataMetadata()
        metadata_dict = literal_eval(self.search_result['extra_information']['value'])
        data_metadata = metadata_dict.pop('data_metadata')
        metadata_all = {"structural_type": d3m_DataFrame,
                        "semantic_types": ["https://metadata.datadrivendiscovery.org/types/Table"],
                        "dimension": {
                            "name": "rows",
                            "semantic_types": ["https://metadata.datadrivendiscovery.org/types/TabularRow"],
                            "length": int(data_metadata['shape_0']),
                            },
                        "schema": "https://metadata.datadrivendiscovery.org/schemas/v0/container.json"
                        }
        return_metadata = return_metadata.update(selector=(), metadata=metadata_all)
        metadata_all_elements = {
            "dimension": {
                "name": "columns",
                "semantic_types": ["https://metadata.datadrivendiscovery.org/types/TabularColumn"],
                "length": int(data_metadata['shape_1']),
            }
        }
        return_metadata = return_metadata.update(selector=(ALL_ELEMENTS,), metadata=metadata_all_elements)

        for each_key, each_value in metadata_dict.items():
            if each_key[:12] == 'column_meta_':
                each_metadata = {
                    "name": each_value['name'],
                    "structural_type": str,
                    "semantic_types": each_value['semantic_type'],
                }
                i = int(each_key.split("_")[-1])
                return_metadata = return_metadata.update(selector=(ALL_ELEMENTS, i), metadata=each_metadata)
        return return_metadata

    def display(self) -> pd.DataFrame:
        """
        function used to see what found inside this search result class in a human vision
        :return: a pandas DataFrame
        """
        if self.search_type == "wikidata":
            column_names = []
            for each in self.search_result["p_nodes_needed"]:
                each_name = self._get_node_name(each)
                column_names.append(each_name)
            column_names = ", ".join(column_names)
            required_variable = []
            required_variable.append(self.search_result["target_q_node_column_name"])
            result = pd.DataFrame({"title": "wikidata search result for " \
                                           + self.search_result["target_q_node_column_name"], \
                                   "columns": column_names, "join columns": required_variable, "score": self._score}, index=[0])

        elif self.search_type == "general":
            title = self.search_result['title']['value']
            column_names = self.search_result['keywords']['value']
            join_columns = self.search_result['variableName']['value']

            result = pd.DataFrame({"title": title, "columns": column_names, "join columns": join_columns, "score": self._score}, index=[0])

        else:
            raise ValueError("Unknown search type with " + self.search_type)
        return result

    def download(self, supplied_data: typing.Union[d3m_Dataset, d3m_DataFrame]=None, connection_url: str = None, generate_metadata=True, return_format="ds") -> container.Dataset:
        """
        Produces a D3M dataset (data plus metadata) corresponding to the search result.
        Every time the download method is called on a search result, it will produce the exact same columns
        (as specified in the metadata -- get_metadata), but the set of rows may depend on the supplied_data.
        Datamart is encouraged to return a dataset that joins well with the supplied data, e.g., has rows that match
        the entities in the supplied data. Datamarts may ignore the supplied_data and return the same data regardless.

        If the supplied_data is None, Datamarts may return None or a default dataset, based on the search query.

        Parameters
        ---------
        supplied_data : container.Dataset
            A D3M dataset containing the dataset that is the target for augmentation. Datamart will try to download data
            that augments the supplied data well.
        connection_url : str
            A connection string used to connect to a specific Datamart deployment. If not provided, the one provided to
            the `Datamart` constructor is used.
        """
        if connection_url:
            self.connection_url = connection_url
        if self.search_type=="general":
            return_df = self.download_general(supplied_data, generate_metadata, return_format)
        elif self.search_type=="wikidata":
            return_df = self.download_wikidata(supplied_data, generate_metadata, return_format)
        else:
            raise ValueError("Unknown search type with " + self.search_type)

        return return_df

    def download_general(self, supplied_data: typing.Union[d3m_Dataset, d3m_DataFrame]=None, generate_metadata=True,
                         return_format="ds", augment_resource_id=AUGMENT_RESOURCE_ID) -> typing.Union[d3m_Dataset, d3m_DataFrame]:
        """
        Specified download function for general datamart Datasets
        :param supplied_data: given supplied data
        :param generate_metadata: whether need to genreate the metadata or not
        :param return_format: set the required return format: d3m_Dataset or d3m_DataFrame
        :param augment_resource_id: the name of the output dataset's resource id, the default is "augmentData"
        :return: a dataset or a dataframe depending on the input
        """
        res_id = None
        if type(supplied_data) is d3m_Dataset:
            res_id, supplied_dataframe = d3m_utils.get_tabular_resource(dataset=supplied_data, resource_id=None)
        elif type(supplied_data) is d3m_DataFrame:
            supplied_dataframe = supplied_data
        else:
            supplied_dataframe = self.supplied_dataframe

        join_pairs_result = []
        candidate_join_column_scores = []

        # start finding pairs
        left_df = copy.deepcopy(supplied_dataframe)
        if self.right_df is None:
            self.right_df = Utils.materialize(metadata=self.search_result)
            right_df = self.right_df
        else:
            print("[INFO] Find downloaded data from previous time, will use that.")
            right_df = self.right_df
        
        left_metadata = Utils.generate_metadata_from_dataframe(data=left_df, original_meta=None)
        right_metadata = Utils.generate_metadata_from_dataframe(data=right_df, original_meta=None)

        if self.join_pairs is None:
            candidate_join_column_pairs = self.get_join_hints(left_df=left_df, right_df=right_df, left_df_src_id=res_id)
        else:
            candidate_join_column_pairs = self.join_pairs
        if len(candidate_join_column_pairs.left_columns) > 1:
            print("[WARN]: multiple joining column pairs found")

        pairs = candidate_join_column_pairs.get_column_number_pairs()
        # generate the pairs for each join_column_pairs
        for each_pair in pairs:
            left_columns = each_pair[0]
            right_columns = each_pair[1]
            try:
                # Only profile the joining columns, otherwise it will be too slow:
                left_metadata = Utils.calculate_dsbox_features(data=left_df, metadata=left_metadata,
                                                               selected_columns=set(left_columns))

                right_metadata = Utils.calculate_dsbox_features(data=right_df, metadata=right_metadata,
                                                                selected_columns=set(right_columns))
                # update with implicit_variable on the user supplied dataset
                if left_metadata.get('implicit_variables'):
                    Utils.append_columns_for_implicit_variables_and_add_meta(left_metadata, left_df)

                print(" - start getting pairs for", str(each_pair))
                right_df_copy = copy.deepcopy(right_df)
                result, self.pairs = RLTKJoiner.find_pair(left_df=left_df, right_df=right_df_copy,
                                                          left_columns=[left_columns], right_columns=[right_columns],
                                                          left_metadata=left_metadata, right_metadata=right_metadata)

                join_pairs_result.append(result)
                # TODO: figure out some way to compute the joining quality
                candidate_join_column_scores.append(1)
            except:
                print("failed when getting pairs for", each_pair)
                traceback.print_exc()

        # choose the best joining results
        all_results = []
        for i in range(len(join_pairs_result)):
            each_result = (pairs[i], candidate_join_column_scores[i], join_pairs_result[i])
            all_results.append(each_result)

        all_results.sort(key=lambda x: x[1], reverse=True)
        if len(all_results) == 0:
            raise ValueError("[ERROR] Failed to get pairs!")

        if return_format == "ds":
            return_df = d3m_DataFrame(all_results[0][2], generate_metadata=False)
            resources = {augment_resource_id: return_df}
            return_result = d3m_Dataset(resources=resources, generate_metadata=False)
            if generate_metadata:
                metadata_shape_part_dict = self._generate_metadata_shape_part(value=return_result, selector=(), supplied_data=supplied_data)
                for each_selector, each_metadata in metadata_shape_part_dict.items():
                    return_result.metadata = return_result.metadata.update(selector=each_selector, metadata=each_metadata)
                return_result.metadata = self._generate_metadata_column_part_for_general(return_result, return_result.metadata, return_format, augment_resource_id)

        elif return_format == "df":
            return_result = d3m_DataFrame(all_results[0][2], generate_metadata=False)
            if generate_metadata:
                metadata_shape_part_dict = self._generate_metadata_shape_part(value=return_result, selector=(), supplied_data=supplied_data)
                for each_selector, each_metadata in metadata_shape_part_dict.items():
                    return_result.metadata = return_result.metadata.update(selector=each_selector, metadata=each_metadata)
                return_result.metadata = self._generate_metadata_column_part_for_general(return_result, return_result.metadata, return_format, augment_resource_id=None)

        else:
            raise ValueError("Invalid return format was given")

        return return_result

    def _generate_metadata_shape_part(self, value, selector, supplied_data=None) -> dict:
        """
        recursively generate all metadata for shape part, return a dict
        :param value:
        :param selector:
        :return:
        """
        if supplied_data is None:
            supplied_data = self.supplied_data
        generated_metadata = dict()
        generated_metadata['schema'] = CONTAINER_SCHEMA_VERSION
        if isinstance(value, d3m_Dataset):  # type: ignore
            generated_metadata['id'] = supplied_data.metadata.query(())['id']
            generated_metadata['name'] = supplied_data.metadata.query(())['name']
            generated_metadata['location_uris'] = supplied_data.metadata.query(())['location_uris']
            generated_metadata['digest'] = supplied_data.metadata.query(())['digest']
            generated_metadata['description'] = supplied_data.metadata.query(())['description']
            generated_metadata['source'] = supplied_data.metadata.query(())['source']
            generated_metadata['version'] = supplied_data.metadata.query(())['version']
            generated_metadata['structural_type'] = supplied_data.metadata.query(())['structural_type']
            generated_metadata['dimension'] = {
                'name': 'resources',
                'semantic_types': ['https://metadata.datadrivendiscovery.org/types/DatasetResource'],
                'length': len(value),
            }

            metadata_dict = collections.OrderedDict([(selector, generated_metadata)])

            for k, v in value.items():
                metadata_dict.update(self._generate_metadata_shape_part(v, selector + (k,)))

            # It is unlikely that metadata is equal across dataset resources, so we do not try to compact metadata here.

            return metadata_dict

        if isinstance(value, d3m_DataFrame):  # type: ignore
            generated_metadata['semantic_types'] = ['https://metadata.datadrivendiscovery.org/types/Table']
            generated_metadata['structural_type'] = d3m_DataFrame
            generated_metadata['dimension'] = {
                'name': 'rows',
                'semantic_types': ['https://metadata.datadrivendiscovery.org/types/TabularRow'],
                'length': value.shape[0],
            }

            metadata_dict = collections.OrderedDict([(selector, generated_metadata)])

            # Reusing the variable for next dimension.
            generated_metadata = {
                'dimension': {
                    'name': 'columns',
                    'semantic_types': ['https://metadata.datadrivendiscovery.org/types/TabularColumn'],
                    'length': value.shape[1],
                },
            }

            selector_all_rows = selector + (ALL_ELEMENTS,)
            metadata_dict[selector_all_rows] = generated_metadata
            return metadata_dict

    def _generate_metadata_column_part_for_general(self, data, metadata_return, return_format, augment_resource_id) -> DataMetadata:
        """
        Inner function used to generate metadata for general search
        """
        # part for adding each column's metadata
        for i, each_column in enumerate(data[augment_resource_id]):
            if return_format == "ds":
                metadata_selector = (augment_resource_id, ALL_ELEMENTS, i)
            elif return_format == "df":
                metadata_selector = (ALL_ELEMENTS, i)
            structural_type = data[augment_resource_id][each_column].dtype.name
            if "int" in structural_type:
                structural_type = int
            elif "float" in structural_type:
                structural_type = float
            else:
                structural_type = str
            metadata_each_column = {"name": each_column, "structural_type": structural_type,
                                    'semantic_types': ('https://metadata.datadrivendiscovery.org/types/Attribute',)}
            metadata_return = metadata_return.update(metadata=metadata_each_column, selector=metadata_selector)

        if return_format == "ds":
            metadata_selector = (augment_resource_id, ALL_ELEMENTS, i + 1)
        elif return_format == "df":
            metadata_selector = (ALL_ELEMENTS, i + 1)
        metadata_joining_pairs = {"name": "joining_pairs", "structural_type": typing.List[int],
                                  'semantic_types': ("http://schema.org/Integer",)}
        metadata_return = metadata_return.update(metadata=metadata_joining_pairs, selector=metadata_selector)

        return metadata_return

    def download_wikidata(self, supplied_data: typing.Union[d3m_Dataset, d3m_DataFrame], generate_metadata=True, return_format="ds",augment_resource_id=AUGMENT_RESOURCE_ID) -> typing.Union[d3m_Dataset, d3m_DataFrame]:
        """
        :param supplied_data: input DataFrame
        :param generate_metadata: control whether to automatically generate metadata of the return DataFrame or not
        :return: return_df: the materialized wikidata d3m_DataFrame,
                            with corresponding pairing information to original_data at last column
        """
        # prepare the query
        p_nodes_needed = self.search_result["p_nodes_needed"]
        target_q_node_column_name = self.search_result["target_q_node_column_name"]
        if type(supplied_data) is d3m_DataFrame:
            self.supplied_dataframe = copy.deepcopy(supplied_data)
            self.supplied_data = supplied_data
        elif type(supplied_data) is d3m_Dataset:
            self._res_id, supplied_dataframe = d3m_utils.get_tabular_resource(dataset=supplied_data,
                                                                                  resource_id=None)
            self.supplied_dataframe = copy.deepcopy(supplied_dataframe)
            self.supplied_data = supplied_data

        q_node_column_number = self.supplied_dataframe.columns.tolist().index(target_q_node_column_name)
        q_nodes_list = set(self.supplied_dataframe.iloc[:, q_node_column_number].tolist())
        q_nodes_query = ""
        p_nodes_query_part = ""
        p_nodes_optional_part = ""
        special_request_part = ""
        # q_nodes_list = q_nodes_list[:30]
        for each in q_nodes_list:
            if each != "N/A":
                q_nodes_query += "(wd:" + each + ") \n"
        for each in p_nodes_needed:
            if each not in P_NODE_IGNORE_LIST:
                p_nodes_query_part += " ?" + each
                p_nodes_optional_part += "  OPTIONAL { ?q wdt:" + each + " ?" + each + "}\n"
            if each in SPECIAL_REQUEST_FOR_P_NODE:
                special_request_part += SPECIAL_REQUEST_FOR_P_NODE[each] + "\n"

        sparql_query = "SELECT DISTINCT ?q " + p_nodes_query_part + \
                       " \nWHERE \n{\n  VALUES (?q) { \n " + q_nodes_query + "}\n" + \
                       p_nodes_optional_part + special_request_part + "}\n"

        # if not self.connection_url:
        #     self.connection_url = WIKIDATA_QUERY_SERVER
        #     print("[INFO] Using default connection url: " + self.connection_url)
        # else:
        #     print("[INFO] User-defined connection url given as: " + self.connection_url)
        return_df = d3m_DataFrame()
        try:
            sparql = SPARQLWrapper(self.connection_url)
            sparql.setQuery(sparql_query)
            sparql.setReturnFormat(JSON)
            sparql.setMethod(POST)

            sparql.setRequestMethod(URLENCODED)
            results = sparql.query().convert()
        except:
            print("Getting query of wiki data failed!")
            return return_df

        semantic_types_dict = {
            "q_node": ("http://schema.org/Text", 'https://metadata.datadrivendiscovery.org/types/PrimaryKey')}
        q_node_name_appeared = set()
        for result in results["results"]["bindings"]:
            each_result = {}
            q_node_name = result.pop("q")["value"].split("/")[-1]
            if q_node_name in q_node_name_appeared:
                continue
            q_node_name_appeared.add(q_node_name)
            each_result["q_node"] = q_node_name
            for p_name, p_val in result.items():
                each_result[p_name] = p_val["value"]
                # only do this part if generate_metadata is required
                if p_name not in semantic_types_dict:
                    if "datatype" in p_val.keys():
                        semantic_types_dict[p_name] = (
                            self._get_semantic_type(p_val["datatype"]),
                            'https://metadata.datadrivendiscovery.org/types/Attribute',
                            AUGMENTED_COLUMN_SEMANTIC_TYPE
                        )
                    else:
                        semantic_types_dict[p_name] = (
                            "http://schema.org/Text",
                            'https://metadata.datadrivendiscovery.org/types/Attribute',
                            AUGMENTED_COLUMN_SEMANTIC_TYPE
                        )

            return_df = return_df.append(each_result, ignore_index=True)

        p_name_dict = {"q_node": "q_node"}
        for each in return_df.columns.tolist():
            if each.lower().startswith("p") or each.lower().startswith("c"):
                p_name_dict[each] = self._get_node_name(each) + "_for_" + target_q_node_column_name

        # use rltk joiner to find the joining pairs
        joiner = RLTKJoiner_new()
        joiner.set_join_target_column_names((self.supplied_dataframe.columns[q_node_column_number], "q_node"))
        result, self.pairs = joiner.find_pair(left_df=self.supplied_dataframe, right_df=return_df)

        # if this condition is true, it means "id" column was added which should not be here
        if return_df.shape[1] == len(p_name_dict) + 2 and "id" in return_df.columns:
            return_df = return_df.drop(columns=["id"])

        metadata_new = DataMetadata()
        self.metadata = {}

        # add remained attributes metadata
        for each_column in range(0, return_df.shape[1] - 1):
            current_column_name = p_name_dict[return_df.columns[each_column]]
            metadata_selector = (ALL_ELEMENTS, each_column)
            # here we do not modify the original data, we just add an extra "expected_semantic_types" to metadata
            metadata_each_column = {"name": current_column_name, "structural_type": str,
                                    'semantic_types': semantic_types_dict[return_df.columns[each_column]]}
            self.metadata[current_column_name] = metadata_each_column
            if generate_metadata:
                metadata_new = metadata_new.update(metadata=metadata_each_column, selector=metadata_selector)

        # special for joining_pairs column
        metadata_selector = (ALL_ELEMENTS, return_df.shape[1])
        metadata_joining_pairs = {"name": "joining_pairs", "structural_type": typing.List[int],
                                  'semantic_types': ("http://schema.org/Integer",)}
        if generate_metadata:
            metadata_new = metadata_new.update(metadata=metadata_joining_pairs, selector=metadata_selector)

        # start adding shape metadata for dataset
        if return_format == "ds":
            return_df = d3m_DataFrame(return_df, generate_metadata=False)
            return_df = return_df.rename(columns=p_name_dict)
            resources = {augment_resource_id: return_df}
            return_result = d3m_Dataset(resources=resources, generate_metadata=False)
            if generate_metadata:
                return_result.metadata = metadata_new
                metadata_shape_part_dict = self._generate_metadata_shape_part(value=return_result, selector=(),
                                                                              supplied_data=self.supplied_data)
                for each_selector, each_metadata in metadata_shape_part_dict.items():
                    return_result.metadata = return_result.metadata.update(selector=each_selector,
                                                                           metadata=each_metadata)
            # update column names to be property names instead of number

        elif return_format == "df":
            return_result = d3m_DataFrame(return_df, generate_metadata=False)
            return_result = return_result.rename(columns=p_name_dict)
            if generate_metadata:
                return_result.metadata = metadata_new
                metadata_shape_part_dict = self._generate_metadata_shape_part(value=return_result, selector=(),
                                                                              supplied_data=self.supplied_data)
                for each_selector, each_metadata in metadata_shape_part_dict.items():
                    return_result.metadata = return_result.metadata.update(selector=each_selector,
                                                                           metadata=each_metadata)
        return return_result

    def _get_node_name(self, node_code):
        """
        Inner function used to get the properties(P nodes) names with given P node
        :param node_code: a str indicate the P node (e.g. "P123")
        :return: a str indicate the P node label (e.g. "inception")
        """
        sparql_query = "SELECT DISTINCT ?x WHERE \n { \n" + \
                       "wd:" + node_code + " rdfs:label ?x .\n FILTER(LANG(?x) = 'en') \n} "
        try:
            sparql = SPARQLWrapper(WIKIDATA_QUERY_SERVER)
            sparql.setQuery(sparql_query)
            sparql.setReturnFormat(JSON)
            sparql.setMethod(POST)
            sparql.setRequestMethod(URLENCODED)
            results = sparql.query().convert()
            return results['results']['bindings'][0]['x']['value']
        except:
            print("Getting name of node " + node_code + " failed!")
            return node_code

    def _get_semantic_type(self, datatype: str):
        """
        Inner function used to transfer the wikidata semantic type to D3M semantic type
        :param datatype: a str indicate the semantic type adapted from wikidata
        :return: a str indicate the semantic type for D3M
        """
        special_type_dict = {"http://www.w3.org/2001/XMLSchema#dateTime": "http://schema.org/DateTime",
                             "http://www.w3.org/2001/XMLSchema#decimal": "http://schema.org/Float",
                             "http://www.opengis.net/ont/geosparql#wktLiteral": "https://metadata.datadrivendiscovery.org/types/Location"}
        default_type = "http://schema.org/Text"
        if datatype in special_type_dict:
            return special_type_dict[datatype]
        else:
            print("not seen type : ", datatype)
            return default_type

    def _do_wikifier(self, supplied_data):
        if type(supplied_data) is d3m_Dataset:
            self._res_id, _ = d3m_utils.get_tabular_resource(dataset=supplied_data, resource_id=None, has_hyperparameter=False)
            supplied_data_df = supplied_data[self._res_id]
        elif type(supplied_data) is d3m_DataFrame:
            supplied_data_df = supplied_data
        import wikifier
        all_columns = list(range(supplied_data_df.shape[1]))
        can_wikifier_columns = []
        for each in all_columns:
            if type(supplied_data) is d3m_Dataset:
                selector = (self._res_id, ALL_ELEMENTS, each)
            elif type(supplied_data) is d3m_DataFrame:
                selector = (ALL_ELEMENTS, each)
            each_column_semantic_type = supplied_data.metadata.query(selector)['semantic_types']
            if 'http://schema.org/Integer' not in each_column_semantic_type and 'http://schema.org/Float' not in each_column_semantic_type:
                can_wikifier_columns.append(each)

        output_ds = copy.deepcopy(supplied_data)
        wikifier_res = wikifier.produce(pd.DataFrame(supplied_data_df), can_wikifier_columns, None)
        output_ds[self._res_id] = d3m_DataFrame(wikifier_res, generate_metadata=False)
        # update metadata on column length
        selector = (self._res_id, ALL_ELEMENTS)
        old_meta = dict(output_ds.metadata.query(selector))
        old_meta_dimension = dict(old_meta['dimension'])
        old_column_length = old_meta_dimension['length']
        old_meta_dimension['length'] = wikifier_res.shape[1]
        old_meta['dimension'] = frozendict.FrozenOrderedDict(old_meta_dimension)
        new_meta = frozendict.FrozenOrderedDict(old_meta)
        output_ds.metadata = output_ds.metadata.update(selector, new_meta)

        # update each column's metadata
        for i in range(old_column_length, wikifier_res.shape[1]):
            selector = (self._res_id, ALL_ELEMENTS, i)
            metadata = {"name": wikifier_res.columns[i],
                        "structural_type": str,
                        'semantic_types': (
                            "http://schema.org/Text",
                            "https://metadata.datadrivendiscovery.org/types/CategoricalData",
                            "https://metadata.datadrivendiscovery.org/types/Attribute",
                            "http://wikidata.org/qnode"
                        )}
            output_ds.metadata = output_ds.metadata.update(selector, metadata)
        return output_ds

    def augment(self, supplied_data, augment_columns=None, connection_url: str = None):
        """
        download and join using the TabularJoinSpec from get_join_hints()
        """
        if connection_url:
            self.connection_url = connection_url

        if self.search_type == "wikifier":
            result = self._do_wikifier(supplied_data)
            return result
        if type(supplied_data) is d3m_DataFrame:
            res = self._augment(supplied_data=supplied_data, generate_metadata=True, return_format="df", augment_resource_id=AUGMENT_RESOURCE_ID)
        elif type(supplied_data) is d3m_Dataset:
            self._res_id, self.supplied_data = d3m_utils.get_tabular_resource(dataset=supplied_data, resource_id=None, has_hyperparameter=False)
            res = self._augment(supplied_data=supplied_data, generate_metadata=True, return_format="ds", augment_resource_id=AUGMENT_RESOURCE_ID)
        res[AUGMENT_RESOURCE_ID] = res[AUGMENT_RESOURCE_ID].astype(str)
        return res

    def _augment(self, supplied_data, augment_columns=None, generate_metadata=True, return_format="ds", augment_resource_id=AUGMENT_RESOURCE_ID):
        """
        download and join using the TabularJoinSpec from get_join_hints()
        """
        if type(supplied_data) is d3m_Dataset:
            supplied_data_df = supplied_data[self._res_id]
        elif type(supplied_data) is d3m_DataFrame:
            supplied_data_df = supplied_data
        else:
            supplied_data_df = self.supplied_dataframe

        if supplied_data_df is None:
            raise ValueError("Can't find supplied data!")

        download_result = self.download(supplied_data=supplied_data_df, generate_metadata=False, return_format="df")
        download_result = download_result.drop(columns=['joining_pairs'])
        df_joined = pd.DataFrame()
        column_names_to_join = None
        r1_paired = set()
        for r1, r2 in self.pairs:
            r1_int = int(r1)
            if r1_int in r1_paired:
                continue
            r1_paired.add(r1_int)
            left_res = supplied_data_df.loc[r1_int]
            right_res = download_result.loc[int(r2)]
            if column_names_to_join is None:
                column_names_to_join = right_res.index.difference(left_res.index)
                matched_rows = right_res.index.intersection(left_res.index)
                columns_new = left_res.index.tolist()
                columns_new.extend(column_names_to_join.tolist())
            new = pd.concat([left_res, right_res[column_names_to_join]])
            df_joined = df_joined.append(new, ignore_index=True)

        # add up the rows don't have pairs
        unpaired_rows = set(range(supplied_data_df.shape[0])) - r1_paired
        if len(unpaired_rows) > 0:
            df_joined = df_joined.append(supplied_data_df.loc[unpaired_rows], ignore_index=True)

        # ensure that the original dataframe columns are at the first left part
        df_joined = df_joined[columns_new]
        # if search with wikidata, we can remove duplicate Q node column

        if self.search_type == "wikidata":
            df_joined = df_joined.drop(columns=['q_node'])

        if 'id' in df_joined.columns:
            df_joined = df_joined.drop(columns=['id'])

        if generate_metadata:
            # put d3mIndex at first column
            columns_all = list(df_joined.columns)
            if 'd3mIndex' in df_joined.columns:
                oldindex = columns_all.index('d3mIndex')
                columns_all.insert(0, columns_all.pop(oldindex))
            else:
                self._logger.warning("No d3mIndex column found after data-mart augment!!!")
            df_joined = df_joined[columns_all]

        # start adding column metadata for dataset
        if generate_metadata:
            metadata_dict_left = {}
            metadata_dict_right = {}
            if self.search_type == "general":
                for i, each in enumerate(df_joined):
                    # description = each['description']
                    dtype = df_joined[each].dtype.name
                    if "float" in dtype:
                        semantic_types = (
                            "http://schema.org/Float",
                            "https://metadata.datadrivendiscovery.org/types/Attribute",
                            AUGMENTED_COLUMN_SEMANTIC_TYPE
                        )
                    elif "int" in dtype:
                        semantic_types = (
                            "http://schema.org/Integer",
                            "https://metadata.datadrivendiscovery.org/types/Attribute",
                            AUGMENTED_COLUMN_SEMANTIC_TYPE
                        )
                    else:
                        semantic_types = (
                            "https://metadata.datadrivendiscovery.org/types/CategoricalData",
                            "https://metadata.datadrivendiscovery.org/types/Attribute",
                            AUGMENTED_COLUMN_SEMANTIC_TYPE
                        )

                    each_meta = {
                        "name": each,
                        "structural_type": str,
                        "semantic_types": semantic_types,
                        # "description": description
                    }
                    metadata_dict_right[each] = frozendict.FrozenOrderedDict(each_meta)
            else:
                metadata_dict_right = self.metadata

            if return_format == "df":
                try:
                    left_df_column_legth = supplied_data.metadata.query((metadata_base.ALL_ELEMENTS,))['dimension']['length']
                except:
                    traceback.print_exc()
                    raise ValueError("No getting metadata information failed!")
            elif return_format == "ds":
                left_df_column_legth = supplied_data.metadata.query((self._res_id, metadata_base.ALL_ELEMENTS,))['dimension']['length']

            # add the original metadata
            for i in range(left_df_column_legth):
                if return_format == "df":
                    each_selector = (ALL_ELEMENTS, i)
                elif return_format == "ds":
                    each_selector = (self._res_id, ALL_ELEMENTS, i)
                each_column_meta = supplied_data.metadata.query(each_selector)
                metadata_dict_left[each_column_meta['name']] = each_column_meta

            metadata_new = metadata_base.DataMetadata()
            metadata_old = copy.copy(supplied_data.metadata)

            new_column_names_list = list(df_joined.columns)
            # update each column's metadata
            for i, current_column_name in enumerate(new_column_names_list):
                if return_format == "df":
                    each_selector = (metadata_base.ALL_ELEMENTS, i)
                elif return_format == "ds":
                    each_selector = (augment_resource_id, ALL_ELEMENTS, i)
                if current_column_name in metadata_dict_left:
                    new_metadata_i = metadata_dict_left[current_column_name]
                else:
                    new_metadata_i = metadata_dict_right[current_column_name]
                metadata_new = metadata_new.update(each_selector, new_metadata_i)

            # start adding shape metadata for dataset
            if return_format == "ds":
                return_df = d3m_DataFrame(df_joined, generate_metadata=False)
                resources = {augment_resource_id: return_df}
                return_result = d3m_Dataset(resources=resources, generate_metadata=False)
                if generate_metadata:
                    return_result.metadata = metadata_new
                    metadata_shape_part_dict = self._generate_metadata_shape_part(value=return_result, selector=(), supplied_data=supplied_data)
                    for each_selector, each_metadata in metadata_shape_part_dict.items():
                        return_result.metadata = return_result.metadata.update(selector=each_selector,
                                                                               metadata=each_metadata)
            elif return_format == "df":
                return_result = d3m_DataFrame(df_joined, generate_metadata=False)
                if generate_metadata:
                    return_result.metadata = metadata_new
                    metadata_shape_part_dict = self._generate_metadata_shape_part(value=return_result, selector=(), supplied_data=supplied_data)
                    for each_selector, each_metadata in metadata_shape_part_dict.items():
                        return_result.metadata = return_result.metadata.update(selector=each_selector,
                                                                               metadata=each_metadata)

            return return_result

    def score(self) -> float:
        return self._score

    def get_metadata(self) -> DataMetadata:
        return self.d3m_metadata

    def set_join_pairs(self, join_pairs: typing.List["TabularJoinSpec"]) -> None:
        """
        manually set up the join pairs
        :param join_pairs: user specified TabularJoinSpec
        :return:
        """
        self.join_pairs = join_pairs

    def get_join_hints(self, left_df, right_df, left_df_src_id=None, right_src_id=None) -> typing.List["TabularJoinSpec"]:
        """
        Returns hints for joining supplied data with the data that can be downloaded using this search result.
        In the typical scenario, the hints are based on supplied data that was provided when search was called.

        The optional supplied_data argument enables the caller to request recomputation of join hints for specific data.

        :return: a list of join hints. Note that datamart is encouraged to return join hints but not required to do so.
        """
        right_join_column_name = self.search_result['variableName']['value']
        left_columns = []
        right_columns = []

        for each in self.query_json['variables'].keys():
            left_index = left_df.columns.tolist().index(each)
            right_index = right_df.columns.tolist().index(right_join_column_name)
            left_index_column = DatasetColumn(resource_id=left_df_src_id, column_index=left_index)
            right_index_column = DatasetColumn(resource_id=right_src_id, column_index=right_index)
            left_columns.append([left_index_column])
            right_columns.append([right_index_column])
        
        results = TabularJoinSpec(left_columns=left_columns, right_columns=right_columns)

        return results

    def serialize(self):
        result = dict()
        if self.search_type == "general":
            result['id'] = self.search_result['datasetLabel']['value']
            result['score'] = float(self.search_result['score']['value'])
        elif self.search_type == "wikidata":
            result['id'] = "wikidata search on " + str(self.search_result['p_nodes_needed']) + " with column " + \
                           self.search_result['target_q_node_column_name']
            result['score'] = self._score
        else:
            result['id'] = ""
            result['score'] = 0

        result['metadata'] = dict()
        result['metadata']['search_result'] = self.search_result
        result['metadata']['query_json'] = self.query_json
        result['metadata']['search_type'] = self.search_type
        augmentation = dict()
        augmentation['properties'] = "join"
        if self.search_type == "general":
            left_col_number = []
            right_col_number = None
            for each_key, each_value in literal_eval(self.search_result['extra_information']['value']).items():
                if each_value['name'] == self.search_result['variableName']['value']:
                    right_col_number = int(each_key.split("_")[-1])
                    break
            augmentation['right_columns'] = [right_col_number]
            if self.supplied_dataframe is None:
                print("[ERROR] Can't get supplied dataframe information, failed to find the left join column number")
            else:
                for each in self.query_json['variables'].keys():
                    left_col_number.append(self.supplied_dataframe.columns.tolist().index(each))
            augmentation['left_columns'] = left_col_number
        elif self.search_type == "wikidata":
            left_col_number = self.supplied_dataframe.columns.tolist().index(self.search_result['target_q_node_column_name'])
            augmentation['left_columns'] = [left_col_number]
            right_col_number = len(self.search_result['p_nodes_needed']) + 1
            augmentation['right_columns'] = [right_col_number]
        result['augmentation'] = augmentation
        result['datamart_type'] = 'isi'
        result_str = json.dumps(result)

        return result_str

    @classmethod
    def deserialize(cls, serialize_result_str):
        serialize_result = json.loads(serialize_result_str)
        if "datamart_type" not in serialize_result or serialize_result["datamart_type"] != "isi":
            raise ValueError("False datamart type found")
        supplied_data = None  # serialize_result['metadata']['supplied_data']
        search_result = serialize_result['metadata']['search_result']
        query_json = serialize_result['metadata']['query_json']
        search_type = serialize_result['metadata']['search_type']
        return DatamartSearchResult(search_result, supplied_data, query_json, search_type)

    def __getstate__(self) -> typing.Dict:
        """
        This method is used by the pickler as the state of object.
        The object can be recovered through this state uniquely.
        Returns:
            state: Dict
                dictionary of important attributes of the object
        """
        state = {}
        state["search_result"] = self.__dict__["search_result"]
        state["query_json"] = self.__dict__["query_json"]
        state["search_type"] = self.__dict__["search_type"]

        return state

    def __setstate__(self, state: typing.Dict) -> None:
        """
        This method is used for unpickling the object. It takes a dictionary
        of saved state of object and restores the object to that state.
        Args:
            state: typing.Dict
                dictionary of the objects picklable state
        Returns:
        """
        self = self.__init__(search_result=state['search_result'], supplied_data=None, query_json=state['query_json'], search_type=state['search_type'])



class AugmentSpec:
    """
    Abstract class for D3M augmentation specifications
    """
    pass


class TabularJoinSpec(AugmentSpec):
    """
    A join spec specifies a possible way to join a left dataset with a right dataset. The spec assumes that it may
    be necessary to use several columns in each datasets to produce a key or fingerprint that is useful for joining
    datasets. The spec consists of two lists of column identifiers or names (left_columns, left_column_names and
    right_columns, right_column_names).

    In the simplest case, both left and right are singleton lists, and the expectation is that an appropriate
    matching function exists to adequately join the datasets. In some cases equality may be an appropriate matching
    function, and in some cases fuzz matching is required. The join spec does not specify the matching function.

    In more complex cases, one or both left and right lists contain several elements. For example, the left list
    may contain columns for "city", "state" and "country" and the right dataset contains an "address" column. The join
    spec pairs up ["city", "state", "country"] with ["address"], but does not specify how the matching should be done
    e.g., combine the city/state/country columns into a single column, or split the address into several columns.
    """
    def __init__(self, left_columns: typing.List[typing.List[DatasetColumn]], right_columns: typing.List[typing.List[DatasetColumn]], 
                 left_resource_id: str=None, right_resource_id: str=None) -> None:

        self.left_resource_id = left_resource_id
        self.right_resource_id = right_resource_id
        self.left_columns = left_columns
        self.right_columns = right_columns
        if len(self.left_columns) != len(self.right_columns):
            shorter_len = min(len(self.right_columns), len(self.left_columns))
            self.left_columns = self.left_columns[:shorter_len]
            self.right_columns = self.right_columns[:shorter_len]
            print("The join spec length on left and right are different! Part of them will be ignored")

        # we can have list of the joining column pairs
        # each list inside left_columns/right_columns is a candidate joining column for that dataFrame
        # each candidate joining column can also have multiple columns

    def get_column_number_pairs(self):
        """
            A simple function used to get the pairs of column numbers only
            For example, it will return a join pair like ([1,2], [1])
        """
        all_pairs = []
        for each in zip(self.left_columns, self.right_columns):
            left = []
            right = []
            for each_left_col in each[0]:
                left.append(each_left_col.column_index)
            for each_right_col in each[1]:
                right.append(each_right_col.column_index)
            all_pairs.append((left, right))
        return all_pairs


class UnionSpec(AugmentSpec):
    """
    A union spec specifies how to combine rows of a dataframe in the left dataset with a dataframe in the right dataset.
    The dataframe after union should have the same columns as the left dataframe.

    Implementation: TBD
    """
    pass


class TemporalGranularity(utils.Enum):
    YEAR = 1
    MONTH = 2
    DAY = 3
    HOUR = 4
    SECOND = 5


class GeospatialGranularity(utils.Enum):
    COUNTRY = 1
    STATE = 2
    COUNTY = 3
    CITY = 4
    POSTAL_CODE = 5


class ColumnRelationship(utils.Enum):
    CONTAINS = 1
    SIMILAR = 2
    CORRELATED = 3
    ANTI_CORRELATED = 4
    MUTUALLY_INFORMATIVE = 5
    MUTUALLY_UNINFORMATIVE = 6


class DatamartQuery:
    """
    A Datamart query consists of two parts:

    * A list of keywords.

    * A list of required variables. A required variable specifies that a matching dataset must contain a variable
      satisfying the constraints provided in the query. When multiple required variables are given, the matching
      dataset should contain variables that match each of the variable constraints.

    The matching is fuzzy. For example, when a user specifies a required variable spec using named entities, the
    expectation is that a matching dataset contains information about the given named entities. However, due to name,
    spelling, and other differences it is possible that the matching dataset does not contain information about all
    the specified entities.

    In general, Datamart will do a best effort to satisfy the constraints, but may return datasets that only partially
    satisfy the constraints.
    """
    def __init__(self, keywords: typing.List[str]=list(), variables: typing.List['VariableConstraint']=list(),
                 search_type: str="general") -> None:
        self.search_type = search_type
        self.keywords = keywords
        self.variables = variables


class VariableConstraint(object):
    """
    Abstract class for all variable constraints.
    """


class NamedEntityVariable(VariableConstraint):
    """
    Specifies that a matching dataset must contain a variable including the specified set of named entities.

    For example, if the entities are city names, the expectation is that a matching dataset must contain a variable
    (column) with the given city names. Due to spelling differences and incompleteness of datasets, the returned
    results may not contain all the specified entities.

    Parameters
    ----------
    entities : List[str]
        List of strings that should be contained in the matched dataset column.
    """
    def __init__(self, entities: typing.List[str]) -> None:
        self.entities = entities


class TemporalVariable(VariableConstraint):
    """
    Specifies that a matching dataset should contain a variable with temporal information (e.g., dates) satisfying
    the given constraint.

    The goal is to return a dataset that covers the requested temporal interval and includes
    data at a requested level of granularity.

    Datamart will return best effort results, including datasets that may not fully cover the specified temporal
    interval or whose granularity is finer or coarser than the requested granularity.

    Parameters
    ----------
    start : datetime
        A matching dataset should contain a variable with temporal information that starts earlier than the given start.
    end : datetime
        A matching dataset should contain a variable with temporal information that ends after the given end.
    granularity : TemporalGranularity
        A matching dataset should provide temporal information at the requested level of granularity.
    """
    def __init__(self, start: datetime.datetime, end: datetime.datetime, granularity: TemporalGranularity = None) -> None:
        self.start = start
        self.end = end
        self.granularity = granularity


class GeospatialVariable(VariableConstraint):
    """
    Specifies that a matching dataset should contain a variable with geospatial information that covers the given
    bounding box.

    A matching dataset may contain variables with latitude and longitude information (in one or two columns) that
    cover the given bounding box.

    Alternatively, a matching dataset may contain a variable with named entities of the given granularity that provide
    some coverage of the given bounding box. For example, if the bounding box covers a 100 mile square in Southern
    California, and the granularity is City, the result should contain Los Angeles, and other cities in Southern
    California that intersect with the bounding box (e.g., Hawthorne, Torrance, Oxnard).

    Parameters
    ----------
    latitude1 : float
        The latitude of the first point
    longitude1 : float
        The longitude of the first point
    latitude2 : float
        The latitude of the second point
    longitude2 : float
        The longitude of the second point
    granularity : GeospatialGranularity
        Requested geospatial values are well matched with the requested granularity
    """
    def __init__(self, latitude1: float, longitude1: float, latitude2: float, longitude2: float, granularity: GeospatialGranularity = None) -> None:
        self.latitude1 = latitude1
        self.longitude1 = longitude1
        self.latitude2 = latitude2
        self.longitude2 = longitude2
        self.granularity = granularity


class TabularVariable(object):
    """
    Specifies that a matching dataset should contain variables related to given columns in the supplied_dataset.

    The relation ColumnRelationship.CONTAINS specifies that string values in the columns overlap using the string
    equality comparator. If supplied_dataset columns consists of temporal or spatial values, then
    ColumnRelationship.CONTAINS specifies overlap in temporal range or geospatial bounding box, respectively.

    The relation ColumnRelationship.SIMILAR specifies that string values in the columns overlap using fuzzy string matching.

    The relations ColumnRelationship.CORRELATED and ColumnRelationship.ANTI_CORRELATED specify the columns are
    correlated and anti-correlated, respectively.

    The relations ColumnRelationship.MUTUALLY_INFORMATIVE and ColumnRelationship.MUTUALLY_UNINFORMATIVE specify the columns
    are mutually and anti-correlated, respectively.

    Parameters:
    -----------
    columns : typing.List[int]
        Specify columns in the dataframes of the supplied_dataset
    relationship : ColumnRelationship
        Specifies how the the columns in the supplied_dataset are related to the variables in the matching dataset.
    """
    def __init__(self, columns: typing.List[DatasetColumn], relationship: ColumnRelationship) -> None:
        self.columns = columns
        self.relationship = relationship



from datamart_isi.materializers.materializer_base import MaterializerBase
from datamart_isi.materializers.parsers import CSVParser, ExcelParser, HTMLParser, JSONParser
import pandas as pd
import typing
from datamart_isi.materializers.parsers.parse_result import ParseResult


class GeneralMaterializer(MaterializerBase):
    type2parser = {
        'csv': CSVParser,
        'json': JSONParser,
        'xls': ExcelParser,
        'xlsb': ExcelParser,
        'xlsm': ExcelParser,
        'xlsx': ExcelParser,
        'excel': ExcelParser,
        'asp': HTMLParser,
        'html': HTMLParser
    }

    def __init__(self, **kwargs):
        MaterializerBase.__init__(self, **kwargs)

    @classmethod
    def get_parser_type(cls, url, file_type=None):
        _type = 'none'
        if file_type:
            if file_type.endswith('csv'):
                _type = 'csv'
            elif file_type.endswith('json'):
                _type = 'json'
            elif file_type == 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet':
                _type = 'xlsx'
        else:
            _type = url.rstrip('/').rsplit('/', 1)[-1].rsplit('.', 1)[-1]
        return _type

    @classmethod
    def can_parse(cls, url, file_type=None):
        _type = cls.get_parser_type(url, file_type)
        if _type in ['csv', 'json', 'xls', 'xlsb', 'xlsm', 'xlsx', 'excel', 'asp', 'html']:
            return True
        return False

    def parse(self,
              metadata: dict = None,
              constrains: dict = None) -> typing.List[ParseResult]:
        """ API for get a dataframe.
            Args:
                metadata: json schema for data_type
                constrains: include some constrains like date_range, location and so on
        """
        args = metadata['materialization']['arguments']
        url = args['url']
        _type = self.get_parser_type(url, args.get('file_type'))
        parser = self.type2parser.get(_type, HTMLParser)()
        res = parser.parse(url)

        return res

    def get(self,
            metadata: dict = None,
            constrains: dict = None
            ) -> typing.Optional[pd.DataFrame]:
        """ API for get a dataframe.
            Args:
                metadata: json schema for data_type
                constrains: include some constrains like date_range, location and so on
        """
        args = metadata['materialization']['arguments']
        url = args['url']
        index = args.get('index', 0)
        _type = args.get('file_type') or url.rstrip('/').rsplit('/', 1)[-1].rsplit('.', 1)[-1]
        parser = self.type2parser.get(_type, HTMLParser)()
        res = parser.get(url, index)

        return res



from enum import Enum
import pandas as pd
from rltk import Record


class DistributeType(Enum):
    CATEGORICAL = 'categorical'
    TOKEN_CATEGORICAL = 'token_categorical'
    NON_CATEGORICAL = 'non_categorical'


class DataType(Enum):
    DATETIME = 'datetime'
    NUMBER = 'number'
    STRING = 'string'
    OTHER = 'other'


class NumberType(Enum):
    FLOAT = 'float'
    INT = 'int'
    OTHER = 'other'


class StringType(Enum):
    WORD = 'word'
    PHRASE = 'phrase'
    SENTENCE = 'sentence'
    PARAGRAPH = 'paragraph'
    OTHER = 'other'


class DatetimeResolution(Enum):
    YEAR = 'year'
    MONTH = 'month'
    DAY = 'day'
    HOUR = 'hour'
    MINUTE = 'minute'
    SECOND = 'second'
    MILLISECOND = 'millisecond'


class FeatureBase(object):
    """
    One or more columns in a dataframe that groups together as a feature for joining(Entity Resolution) .

    Args:
        dfw: DataFrameWrapper - include the pd.dataframe, the metadata, and the columns used as joining features
        indexes: list[int] - indexes of the columns for this feature, whose values will be merged as a single value \
                 for joining
    """
    def __init__(self, df: pd.DataFrame, indexes, metadata, distribute_type, data_type):
        self._df = df
        self._indexes = indexes
        self._metadata = metadata

        self._multi_column = False if len(indexes) == 1 else True
        self._headers = [str(df.iloc[:, i].name) for i in indexes]
        self._name = '|'.join(self._headers)

        self._distribute_type = distribute_type
        self._data_type = data_type

    @property
    def multi_column(self) -> bool:
        return self._multi_column

    @property
    def metadata(self) -> dict:
        return self._metadata

    @property
    def distribute_type(self) -> DistributeType:
        return self._distribute_type

    @property
    def data_type(self) -> DataType:
        return self._data_type

    @property
    def name(self):
        """

        Returns: str - name of the feature, the joining of columns headers by '|', also the rltk record property name

        """
        return self._name

    def value_merge_func(self, record: Record):
        """
        Define the function to map the multiple columns values to a single feature value. Only useful when \
        self.multi_column is True

        Args:
            record: dict - the rltk record raw_object, which is a dict in {header1: value1, header2: value2 ... } \
                    representing a row of the original dataframe

        Returns: merged value as the feature value for joining

        """
        return [getattr(record, header) for header in self._headers]

    def similarity_functions(self):
        """
        Define which similarity functions should be use when compare two values of this feature, the order is \
        representing the priorities of different similarity functions .

        Returns: list of functions

        """
        def default_similarity_function(r1, r2):
            v1 = self.value_merge_func(r1)
            v2 = self.value_merge_func(r2)
            return 1 if v1 == v2 else 0

        return [default_similarity_function]



def convert_dict(prop_idents) -> dict:
    """
    Transfet the dict generated by sparql.py, store them in the format of:
    {
      "123" : [ "P345/Q123", "P789/Q234" ],
     "453" : [ "P789/Q43", "P256/Q456", "P31/Q567" ],
      ...
    }
    :param prop_idents:
    :return:
    """
    identifier_nodes_dict = {}
    count = 0
    keys = set()
    for P_number,p_entity in prop_idents.items():
        count += 1
        print(P_number,count)
        for identi,Q_number in p_entity.items():
            #if identi in identifier_nodes_dict.keys():
            if identi in keys:
                identifier_nodes_dict[identi].append(P_number + '/' + Q_number)
            else:
                keys.add(identi)
                identifier_nodes_dict[identi] = [P_number + '/' + Q_number]
    return identifier_nodes_dict


def dict_slice(dict_input, start, end) -> dict:
    """
    take slice for python dict
    :param dict_input: a dict for slicing
    :param start: start position for splicing
    :param end: end position for slicing
    :return: the sliced dict
    """
    keys = dict_input.keys()
    dict_slice = {}
    for k in keys[start:end]:
        dict_slice[k] = dict_input[k]
    return dict_slice



from datamart_isi.materializers.parsers.parser_base import *
# try:
#     from etk.extractors.table_extractor import TableExtractor
# except OSError:
#     from spacy.cli import download
#     download('en_core_web_sm')
#     from etk.extractors.table_extractor import TableExtractor


class HTMLParser(ParserBase):

    def get_all(self, url: str) -> typing.List[pd.DataFrame]:
        print("This parser was no longer exist!")
        return [pd.DataFrame()]
        # content = self.load_content(url)
        # table_extractor = TableExtractor()
        # extractions = table_extractor.extract(content.decode('utf-8', 'ignore'))
        # rows = extractions[0].value.get('rows')
        # if not rows:
        #     return []
        # cells = [[cell['text'] for cell in row['cells']] for row in extractions[0].value['rows']]
        # headers = None
        # try:
        #     if not rows[0]['cells'][0]['cell'].startswith('<td>'):
        #         headers = cells.pop(0)
        # except (IndexError, KeyError, SyntaxError, AttributeError):
        #     pass
        # return [pd.DataFrame(cells, columns=headers)]




import pandas as pd
import dateutil.parser
import typing
from datamart_isi.metadata.variable_metadata import VariableMetadata
from datamart_isi.metadata.global_metadata import GlobalMetadata
from pandas.api.types import is_object_dtype


class BasicProfiler(object):
    def __init__(self):
        pass

    @staticmethod
    def profile_named_entity(column: pd.Series) -> typing.List[str]:
        """Profiling this named entities column, use when this column is marked as a named entities column.

        Args:
            column: pandas Series column.

        Returns:
            list of named entities string
        """
        return column.dropna().unique().astype(str).tolist()

    @staticmethod
    def named_entity_column_recognize(column: pd.Series) -> bool:
        """Ideally run a NER on the column and profile.

        Args:
            column: pandas Series column.

        Returns:
            True if this column is a named entity column
            False if not a named entity column
        """
        # TODO: Write a real NER here maybe

        if is_object_dtype(column):
            try:
                pd.to_datetime(column)
                return False
            except:
                all_ = column.dropna()
                nums = pd.to_numeric(all_, errors='coerce').dropna()
                all_ = all_.unique()
                if len(all_) == 0 or len(nums) / len(all_) > 0.5:
                    return False
                return True
        return False

    @staticmethod
    def profile_semantic_type(column: pd.Series) -> typing.List:
        # TODO: we need to check str is text or categorical here
        # when to use "https://metadata.datadrivendiscovery.org/types/CategoricalData"
        semantic_types = ["https://metadata.datadrivendiscovery.org/types/Attribute"]
        if column.dtype.name == "object":
            semantic_types.append("http://schema.org/Text")
        elif "float" in column.dtype.name:
            semantic_types.append("http://schema.org/Float")
        elif "int" in column.dtype.name:
            semantic_types.append("http://schema.org/Integer")
        return semantic_types

    @staticmethod
    def profile_temporal_coverage(column: pd.Series, coverage: dict = None) -> typing.Union[dict, bool]:
        """Profiling this temporal column .

        Args:
            coverage: temporal coverage dict
            column: pandas Series column.

        Returns:
            temporal coverage dict
            False if there is no temporal related element detected
        """

        temporal_count = 0
        min_datetime = None
        max_datetime = None
        if len(column) == 0:
            return {
                "start": None,
                "end": None
            }

        for element in column:
            # TODO: now pure time will be treat as "today"'s time, and generate a range
            # TODO: improve date/time detection and parse
            try:
                if isinstance(element, str):
                    this_datetime = dateutil.parser.parse(element)
                else:
                    if len(str(element)) == 4 and column.name.lower() == 'year':
                        this_datetime = dateutil.parser.parse(str(element))
                    else:
                        break
                temporal_count += 1
                if not min_datetime:
                    min_datetime = this_datetime
                min_datetime = min(min_datetime, this_datetime)
                if not max_datetime:
                    max_datetime = this_datetime
                max_datetime = max(max_datetime, this_datetime)
            except:
                pass

        if not min_datetime and not max_datetime:
            return False

        if temporal_count < len(column) // 2:
            return False

        if not coverage:
            coverage = {
                "start": None,
                "end": None
            }

        if not coverage['start']:
            coverage['start'] = min_datetime.isoformat()
        if not coverage['end']:
            coverage['end'] = max_datetime.isoformat()
        return coverage

    @staticmethod
    def construct_variable_description(column: pd.Series) -> str:
        """construct description for this variable .

        Args:
            column: pandas Series column.

        Returns:
            string of description
        """

        ret = "column name: {}, dtype: {}".format(
            column.name, column.dtype
        )
        return ret

    @staticmethod
    def construct_global_title(data: pd.DataFrame) -> str:
        """construct title for this dataset .

        Args:
            data: pandas dataframe.

        Returns:
            string of title
        """

        return " ".join([str(_) for _ in data.columns.tolist()])

    @staticmethod
    def construct_global_description(data: pd.DataFrame) -> str:
        """construct description for this dataset .

        Args:
            data: pandas dataframe.

        Returns:
            string of description
        """

        return ", ".join(["{} : {}".format(data.columns[i], data.dtypes[i]) for i in range(data.shape[1])])

    @staticmethod
    def construct_global_keywords(data: pd.DataFrame) -> str:
        """construct keywords for this dataset .

        Args:
            data: pandas dataframe.

        Returns:
            list of keywords
        """

        return data.columns.astype(str).tolist()

    @classmethod
    def basic_profiling_column(cls,
                               description: dict,
                               variable_metadata: VariableMetadata,
                               column: pd.Series
                               ) -> VariableMetadata:
        """Profiling single column for necessary fields of metadata, if data is present .

        Args:
            description: description dict about the column.
            variable_metadata: the original VariableMetadata instance.
            column: the column to profile.

        Returns:
            profiled VariableMetadata instance
        """

        if not variable_metadata.name:
            variable_metadata.name = str(column.name)

        if not variable_metadata.description:
            variable_metadata.description = cls.construct_variable_description(column)

        if variable_metadata.named_entity is None:
            variable_metadata.named_entity = cls.profile_named_entity(column)

        elif variable_metadata.named_entity is False and not description:
            if cls.named_entity_column_recognize(column):
                variable_metadata.named_entity = cls.profile_named_entity(column)

        if variable_metadata.temporal_coverage is not False:
            if not variable_metadata.temporal_coverage['start'] or not variable_metadata.temporal_coverage['end']:
                variable_metadata.temporal_coverage = cls.profile_temporal_coverage(
                    column=column, coverage=variable_metadata.temporal_coverage)
        elif not description:
            temporal_coverage = cls.profile_temporal_coverage(column=column)
            if temporal_coverage:
                variable_metadata.temporal_coverage = temporal_coverage

        if not variable_metadata.semantic_type:
            variable_metadata.semantic_type = cls.profile_semantic_type(column)

        return variable_metadata

    @classmethod
    def basic_profiling_entire(cls, global_metadata: GlobalMetadata, data: pd.DataFrame) -> GlobalMetadata:
        """Profiling entire dataset for necessary fields of metadata, if data is present .

        Args:
            global_metadata: the original GlobalMetadata instance.
            data: dataframe of data.

        Returns:
            profiled GlobalMetadata instance
        """

        if not global_metadata.title:
            global_metadata.title = cls.construct_global_title(data)

        if not global_metadata.description:
            global_metadata.description = cls.construct_global_description(data)

        if not global_metadata.keywords:
            global_metadata.keywords = cls.construct_global_keywords(data)

        return global_metadata



import json
#write files
b = {} #some dictionary
f1 = open("identifier_nodes_v2_small.json", "w+")
js = json.dumps(b, indent=4)
f1.write(js)
f1.close()

#read files
with open('identifier_nodes_v2_small.json') as f4:
	d = json.load(f4)


import sys, os
import warnings
from dsbox.datapreprocessing.cleaner.wikifier import WikifierHyperparams ,Wikifier
from d3m.metadata.base import Metadata, DataMetadata, ALL_ELEMENTS
from d3m.container import List
from d3m.base import utils as d3m_utils
from datamart import upload
from datamart.utilities.utils import SEARCH_URL, PRODUCTION_ES_INDEX, TEST_ES_INDEX
import numpy as np
import logging
import pandas as pd
import copy
import subprocess
from .metadata_wikipedia import metadata_wikipedia
from elasticsearch import Elasticsearch
import config

warnings.filterwarnings("ignore")
logger = logging.getLogger(__name__)
level = logging.getLevelName('ERROR')
logger.setLevel(level)
Q_NODE_SEMANTIC_TYPE = "http://wikidata.org/qnode"
WIKIDATA_URL = "https://tools.wmflabs.org/sqid/#/view?id="
COLOR_BANK = ['#FFB567', '#36DBFF', '#C1FE9B', '#B89E9E', '#F3FF6D']
ALL_STATES = set(['alabama', 'alaska', 'arizona', 'arkansas', 'california', 'colorado', 'connecticut', 'delaware', 'florida', 'georgia', 'hawaii', 'idaho', 'illinois', 'indiana', 'iowa', 'kansas', 'kentucky', 'louisiana', 'maine', 'maryland', 'massachusetts', 'michigan', 'minnesota', 'mississippi', 'missouri', 'montana', 'nebraska', 'nevada', 'new-hampshire', 'new-jersey', 'new-mexico', 'new-york', 'north-carolina', 'north-dakota', 'ohio', 'oklahoma', 'oregon', 'pennsylvania', 'rhode-island', 'south-carolina', 'south-dakota', 'tennessee', 'texas', 'utah', 'vermont', 'virginia', 'washington', 'west-virginia', 'wisconsin', 'wyoming'])
URL = "http://dsbox02.isi.edu:9200/" # changing this for other queries
INDEX = "datamart_v2"

class DATAMART_ID:
    datamart_id = ""

class Count:
    count = 0

# Disable
def blockPrint():
    sys.stdout = open(os.devnull, 'w')
    sys.stderr = open(os.devnull, 'w')
# Restore
def enablePrint():
    sys.stdout = sys.__stdout__
    sys.stderr = sys.__stderr__

def load_d3m_dataset(input_ds_loc):
    # blockPrint()
    from d3m.container.dataset import D3MDatasetLoader
    loader = D3MDatasetLoader()
    all_dataset_uri = 'file://{}'.format(input_ds_loc)
    input_ds = loader.load(dataset_uri=all_dataset_uri)
    enablePrint()
    return input_ds

def wikifier_for_d3m_all(input_ds):
    # blockPrint()
    wikifier_hyperparams = WikifierHyperparams.defaults()

    wikifier_hyperparams = wikifier_hyperparams.replace({"use_columns":()})
    wikifier_primitive = Wikifier(hyperparams = wikifier_hyperparams)
    output_ds = wikifier_primitive.produce(inputs = input_ds)
    enablePrint()
    return output_ds

def wikifier_for_d3m_fips(input_ds):
    # blockPrint()
    wikifier_hyperparams = WikifierHyperparams.defaults()
    qnodes = List(["P882"])
    wikifier_hyperparams = wikifier_hyperparams.replace({"use_columns":(1,), "specific_q_nodes":qnodes})
    wikifier_primitive = Wikifier(hyperparams = wikifier_hyperparams)
    output_ds = wikifier_primitive.produce(inputs = input_ds)
    enablePrint()
    return output_ds

def print_search_results(search_result):
    out_df = pd.DataFrame()
    for each in search_result:
        out_df = out_df.append(each.display(), ignore_index=True)
    return out_df

def make_clickable_both(val): 
    if '#' not in val:
        return val
    name, url = val.split('#', 1)
    return f'<a href="{url}">{name}</a>'

def highlight_cols0(s):
    color = COLOR_BANK[0]
    return 'background-color: %s' % color

def highlight_cols1(s):
    color = COLOR_BANK[1]
    return 'background-color: %s' % color

def highlight_cols2(s):
    color = COLOR_BANK[2]
    return 'background-color: %s' % color

def highlight_cols3(s):
    color = COLOR_BANK[3]
    return 'background-color: %s' % color

def highlight_cols4(s):
    color = COLOR_BANK[4]
    return 'background-color: %s' % color

def color_ending_with_wikidata(column_names):
    color_column_list = []
    for each in column_names:
        if "_wikidata" in each:
            color_column_list.append(each)
    return color_column_list

def color_after_q_nodes(column_names):
    color_column_list = []
    afterwards_columns = []
    previous_is_wikidata = False
    for each in column_names:
        if previous_is_wikidata and "_wikidata" not in each:
            afterwards_columns.append(each)

        if "_wikidata" in each:
            previous_is_wikidata = True
            color_column_list.append(each)
    return color_column_list, afterwards_columns

def pretty_print(input_ds, ds_type="", display_length=10):
    output_ds = copy.deepcopy(input_ds)
    res_id, inputs_df = d3m_utils.get_tabular_resource(dataset=output_ds, resource_id=None)

    can_mark_dict = {}
    can_mark_from = {}
    can_mark_list = []
    each_column = list(inputs_df.columns)
    if ds_type=="download":
        inputs_df = inputs_df.sort_values(by=['joining_pairs'],ascending=False)
        color_column_list, afterwards_columns = color_after_q_nodes(each_column)
    if ds_type=="wikifier":
        color_column_list = color_ending_with_wikidata(each_column)
    elif ds_type=="wiki_augment":
        color_column_list, afterwards_columns = color_after_q_nodes(each_column)
    else:
        color_column_list = []

    for i in range(inputs_df.shape[1]):
        selector = (res_id,ALL_ELEMENTS, i)
        meta = output_ds.metadata.query(selector)
        if "semantic_types" in meta and Q_NODE_SEMANTIC_TYPE in meta["semantic_types"]:
            can_mark_dict[each_column[i]] = make_clickable_both
            can_mark_list.append(meta["name"])
            temp = meta["name"].split("_wikidata")[0]
            can_mark_from[each_column[i]] = temp
            can_mark_dict[temp] = make_clickable_both
            
    for idx, rows in inputs_df.iterrows():
        for each in can_mark_list:
            if rows[each] is not np.nan:
                rows[can_mark_from[each]] = str(rows[can_mark_from[each]]) + '#'+ WIKIDATA_URL + str(rows[each])
                rows[each] = str(rows[each]) + '#' + WIKIDATA_URL + str(rows[each])
                inputs_df.at[idx,each] = rows[each]
            else:
                inputs_df.at[idx,each] = str(rows[each])

    res = inputs_df.iloc[:display_length,:]
    res = res.style.format(can_mark_dict).applymap(highlight_cols0, subset=pd.IndexSlice[:, color_column_list])
    if ds_type=="wiki_augment" or ds_type=="download":
        res = res.applymap(highlight_cols1, subset=pd.IndexSlice[:, afterwards_columns])
    return res

def download_FBI_data(states=None, python_path=config.python_path, pypath=config.pypath):
    if states is None:
        states = ALL_STATES
    for each_state in states:
        each_state = each_state.lower()
        command_download = python_path + " " + config.pypath +"wikidata-wikifier/wikifier/wikidata/FBI_Crime_Model.py download " + each_state
        p = subprocess.Popen(command_download, stdout=subprocess.PIPE, shell=True, stderr=subprocess.STDOUT)
        while p.poll() == None:
            out = p.stdout.readline().strip()
            if out:
                print (bytes.decode(out))

def generate_FBI_data(states=None, python_path=config.python_path, pypath=config.pypath):
    if states is None:
        states = ALL_STATES
    for each_state in states:
        each_state = each_state.lower()
        command_generate = python_path + " " + pypath +"wikidata-wikifier/wikifier/wikidata/FBI_Crime_Model.py generate " + each_state
        p = subprocess.Popen(command_generate, stdout=subprocess.PIPE, shell=True, stderr=subprocess.STDOUT)
        while p.poll() == None:
            out = p.stdout.readline().strip()
            if out:
                print (bytes.decode(out))

def upload_FBI_data(states=None, mode=""):
    if states is None:
        states = ALL_STATES

    if mode == "test":
        endpoint = config.endpoint_upload_test
    else:
        endpoint = config.endpoint_upload_main

    for each_state in states:
        each_state = each_state.lower()
        command_add = config.python_path + config.config_upload + endpoint + "--user " + config.user + " --passwd " + config.password + " -f " + each_state + ".ttl"
        command_update_truthy = config.python_path + config.config_update_truthy + endpoint + "--user " + config.user + " --passwd " + config.password + " -f changes_" + each_state + ".tsv"
        
        p = subprocess.Popen(command_add, stdout=subprocess.PIPE, shell=True, stderr=subprocess.STDOUT)
        while p.poll() == None:
            out = p.stdout.readline().strip()
            if out:
                print (bytes.decode(out))

        p = subprocess.Popen(command_update_truthy, stdout=subprocess.PIPE, shell=True, stderr=subprocess.STDOUT)
        while p.poll() == None:
            out = p.stdout.readline().strip()
            if out:
                print (bytes.decode(out))

def clean_FBI_data(mode=""):
    if mode == "test":
        endpoint = config.endpoint_upload_test
    else:
        endpoint = config.endpoint_upload_main
    command_clean = python_path + config_clean_up + endpoint
    subprocess.call(command_clean, stdout=subprocess.PIPE, shell=True, stderr=subprocess.STDOUT)

def upload_wikipedia_data():
    all_metadata = metadata_wikipedia
    succeed = upload(meta_list=[all_metadata], es_index=PRODUCTION_ES_INDEX, deduplicate=False)
    DATAMART_ID.datamart_id = succeed[0]['datamart_id']

def delete_by_IDS(selector, es, index=INDEX):
    for v in selector:
        try:
            es.delete(index=index, doc_type = "_doc", id = v)
        except:
            print("failed to delete", v)

def remove_wikipedia_data(ids):
    if ids is None:
        ids = [DATAMART_ID.datamart_id]
    es = Elasticsearch([URL])
    delete_by_IDS(selector=ids,es = es, index=INDEX)


from pandas.io.json import json_normalize
import json

from datamart_isi.materializers.parsers.parser_base import *


class JSONParser(ParserBase):

    def get_all(self, url: str) -> typing.List[pd.DataFrame]:
        """
        Parses json and returns result

        Params:
            - url: (str)

        Returns:
            - result: (list) 
        """

        data = json.loads(self.load_content(url))

        if type(data) is list:
            data = [self._flatten_dict(x) for x in data]
            return [json_normalize(data)]
        else:
            return [json_normalize(self._flatten_dict(data))]

       
    @staticmethod
    def _flatten_dict(d: dict) -> typing.Dict:
        """
        Flattens a dictionary and returns the flattented dict

        Params:
            - d: (dict) Dictionary to be flattened

        Returns:
            - out: (dict) Flattened dictionary
        """
        out = {}
        def flatten(x, name=''):
            if type(x) is dict:
                for a in x:
                    flatten(x[a], name + a + '.')
            elif type(x) is list:
                i = 0
                for a in x:
                    flatten(a, name + str(i) + '.')
                    i += 1
            else:
                out[name[:-1]] = x  

        flatten(d)
        return out


from datamart_isi.materializers.materializer_base import MaterializerBase
from tablextract import tables
from typing import Optional
from pandas import DataFrame
from collections import OrderedDict


WIKIPEDIA_CSS_FILTER = '#content table:not(.infobox):not(.navbox):not(.navbox-inner):not(.navbox-subgroup):not(.sistersitebox)'

class WikitablesMaterializer(MaterializerBase):
    def __init__(self, **kwargs):
        MaterializerBase.__init__(self, **kwargs)

    def get(self, url:str, constrains: dict = None) -> Optional[DataFrame]:
        """ API for get a dataframe.

            Args:
                metadata: json schema for data_type
                variables:
                constrains: include some constrains like date_range, location and so on
        """
        tabs = tables(url, xpath_filter=None, css_filter=WIKIPEDIA_CSS_FILTER)
        # tabs = tables(metadata['url'], xpath_filter=metadata['xpath'])
        data = OrderedDict()
        results = []
        xpath_filters = []
        if len(tabs) > 0:
            for each in tabs:
                tab = each.record
                data = OrderedDict()
                for col in tab[0].keys():
                    data[col] = [r[col] for r in tab]
                results.append(DataFrame(data))
                xpath_filters.append(each.xpath)
        else:
            print("[WARN] No wikitable found from " + url)
        
        return results, xpath_filters

    def get_one(self, url:str, xpath_filter:str, constrains: dict = None) -> Optional[DataFrame]:
        """ API for get a dataframe.

            Args:
                metadata: json schema for data_type
                variables:
                constrains: include some constrains like date_range, location and so on
        """
        tabs = tables(url, xpath_filter=xpath_filter, css_filter=WIKIPEDIA_CSS_FILTER)
        # tabs = tables(metadata['url'], xpath_filter=metadata['xpath'])
        data = OrderedDict()
        results = []
        if len(tabs) > 0:
            tab = tabs[0].record
            data = OrderedDict()
            for col in tab[0].keys():
                data[col] = [r[col] for r in tab]
            return DataFrame(data)

        else:
            print("[WARN] No wikitable found from " + url)    
            return DataFrame()
        


from datamart_isi.metadata.metadata_base import MetadataBase
from datamart_isi.metadata.variable_metadata import VariableMetadata
from datamart_isi.utilities.utils import Utils
import typing


class GlobalMetadata(MetadataBase):
    def __init__(self, description: dict, datamart_id: typing.Union[int, None] = None) -> None:
        """Init method of GlobalMetadata.

        Args:
            description: description dict.
            datamart_id: unique datamart_id.

        Returns:

        """

        super().__init__()

        if datamart_id and not isinstance(datamart_id, int):
            raise ValueError("datamart id must be integer")

        self._metadata["datamart_id"] = datamart_id
        if "title" in description:
            self._metadata["title"] = description["title"]

        if "description" in description:
            self._metadata["description"] = description["description"]

        if "url" in description:
            self._metadata["url"] = description["url"]

        if "keywords" in description:
            self._metadata["keywords"] = description["keywords"]

        if "date_published" in description:
            self._metadata["date_published"] = description["date_published"]
        if self.date_published:
            self.date_published = Utils.date_validate(self.date_published)

        if "date_updated" in description:
            self._metadata["date_updated"] = description["date_updated"]
        if self.date_updated:
            self.date_updated = Utils.date_validate(self.date_updated)

        if "provenance" in description:
            self._metadata["provenance"] = description["provenance"]

        if "original_identifier" in description:
            self._metadata["original_identifier"] = description["original_identifier"]

        if "implicit_variables" in description:
            self._metadata["implicit_variables"] = description["implicit_variables"]

        if "additional_info" in description:
            self._metadata["additional_info"] = description["additional_info"]

        try:
            self._metadata["materialization"] = description["materialization"]
        except:
            raise ValueError("No materialization found")

        if "python_path" not in self.materialization:
            raise ValueError("No python path found in materialization")

        if "arguments" not in self.materialization:
            self._metadata["materialization"]["arguments"] = None

        self._metadata["variables"] = list()
        self._variables = list()

        if "license" in description:
            self._metadata["license"] = description["license"]

    @classmethod
    def construct_global(cls, description, datamart_id=None) -> 'GlobalMetadata':
        return cls(description, datamart_id)

    def add_variable_metadata(self, variable_metadata: VariableMetadata) -> None:
        """Add a variable_metadata to this golbal metadata instance.

        Args:
            variable_metadata: VariableMetadata instance need to be added.

        Returns:

        """

        self._variables.append(variable_metadata)
        self._metadata["variables"].append(variable_metadata.value)

    @property
    def datamart_id(self):
        return self._metadata["datamart_id"]

    @datamart_id.setter
    def datamart_id(self, value):
        self._metadata["datamart_id"] = value

    @property
    def title(self):
        return self._metadata.get("title", False)

    @title.setter
    def title(self, value):
        self._metadata["title"] = value

    @property
    def description(self):
        return self._metadata.get("description", False)

    @description.setter
    def description(self, value):
        self._metadata["description"] = value

    @property
    def url(self):
        return self._metadata.get("url", False)

    @property
    def keywords(self):
        return self._metadata.get("keywords", False)

    @keywords.setter
    def keywords(self, value):
        self._metadata["keywords"] = value

    @property
    def date_published(self):
        return self._metadata.get("date_published", False)

    @date_published.setter
    def date_published(self, value):
        self._metadata["date_published"] = value

    @property
    def date_updated(self):
        return self._metadata.get("date_updated", False)

    @date_updated.setter
    def date_updated(self, value):
        self._metadata["date_updated"] = value

    @property
    def provenance(self):
        return self._metadata.get("provenance", False)

    @property
    def original_identifier(self):
        return self._metadata.get("original_identifier", False)

    @property
    def implicit_variables(self):
        return self._metadata.get("implicit_variables", False)

    @implicit_variables.setter
    def implicit_variables(self, value):
        self._metadata["implicit_variables"] = value

    @property
    def additional_info(self):
        return self._metadata.get("additional_info", False)

    @additional_info.setter
    def additional_info(self, value):
        self._metadata["additional_info"] = value

    @property
    def materialization(self):
        return self._metadata["materialization"]

    @property
    def variables(self):
        return self._variables

    @property
    def variable_values(self):
        return self._metadata["variables"]

    @property
    def license(self):
        return self._metadata.get("license", False)

    @license.setter
    def license(self, value):
        self._metadata["license"] = value



from collections import Counter
import json
import requests
from SPARQLWrapper import SPARQLWrapper, JSON
import typing


P_blacklist = ["P5736", "P3984"]

class FindIdentity:
    def __init__(self):
        pass

    # input: strings, output dictionary of string and Qnode in wikidate
    @staticmethod
    def get_identtifier(strings):
        id_nodes_dict = FindIdentity.call_redis(strings)
        keys = set(id_nodes_dict.keys())
        output = {}  # key string, value:Q123
        print(len(keys))
        P_list = []
        for s in strings:
            if s in keys:
                P_list.extend([P_Q.split('/')[0] for P_Q in id_nodes_dict[s]])
        P_predict = Counter(P_list).most_common(1)[0][0]  # [('P932', 8), ('P1566', 6), ('P698', 2)]
        for s in strings:
            if s in keys:
                for P_Q in id_nodes_dict[s]:  # for P_Q in P_Qlist
                    P_Q_splited = P_Q.split('/')
                    if P_Q_splited[0] == P_predict:
                        output[s] = P_Q_splited[1]
            else:
                output[s] = ''
        return output

    @staticmethod
    def get_identifier_3(strings:typing.List[str], column_name: str=None, target_p_node: str=None):

        id_nodes_dict = FindIdentity.call_redis(strings)
        result = []
        P_list = []
        appeared_threshold = 0.5

        keys = set(id_nodes_dict.keys())

        for s in strings:
            if s in keys:
                # update 2019.5.31: it seems sometimes duplicate relationship will appeared!
                temp = [P_Q.split('/')[0] for P_Q in set(id_nodes_dict[s])]
                temp1 = list(set(temp))
                P_list.extend(temp1)

        if target_p_node is not None:
            P_predicts = [target_p_node]
            print("User-defined P node is " + P_predicts[0])
        else:
            P_predicts = [x[0] for x in Counter(P_list).most_common(5)]  # [('P932', 8), ('P1566', 6), ('P698', 2)]

            for each in P_blacklist:
                try:
                    P_predicts.remove(each)
                except:
                    pass
            if len(P_predicts) == 0:
                print("[ERROR] No candidate P nodes found for input column : [" + column_name + "]")
                return result
            print("The best matching P node is " + P_predicts[0])

        """
        # use edit distance to find best candidate
        P_edit_distance = {}
        for each in P_predicts:
            P_edit_distance[each] = FindIdentity.min_distance(FindIdentity.get_node_name(each), column_name)
        smallest_dist = 1000
        for key, val in P_edit_distance.items():
            if val < smallest_dist:
                best = key
                smallest_dist = val

        import pdb
        pdb.set_trace()
        # best_predicts =
        """
        best_predicts = [P_predicts[0]]

        # print('Top 3 possible properties:')
        # print(P_predicts)

        for P_predict in best_predicts:
            output = {}  # key string, value:Q123
            for s in strings:
                output[s] = ''
                if s in keys:
                    for P_Q in id_nodes_dict[s]:  # for P_Q in P_Qlist
                        P_Q_splited = P_Q.split('/')
                        if P_Q_splited[0] == P_predict:
                            output[s] = P_Q_splited[1]
                            continue
            result.append((P_predict, output))
        return result

    @staticmethod
    def call_redis(qnodes):
        payload = json.dumps({"ids": qnodes})
        search_headers = {'Accept': 'application/json', 'Content-Type': 'application/json'}
        r = requests.post("http://dsbox02.isi.edu/wikifier/get_identifiers", data=payload, headers=search_headers)
        if r:
            val = r.json()
            return val
        else:
            return {}

    @staticmethod
    def get_node_name(node_code):
        """
        Inner function used to get the properties(P nodes) names with given P node
        :param node_code: a str indicate the P node (e.g. "P123")
        :return: a str indicate the P node label (e.g. "inception")
        """
        sparql_query = "SELECT DISTINCT ?x WHERE \n { \n" + \
                       "wd:" + node_code + " rdfs:label ?x .\n FILTER(LANG(?x) = 'en') \n} "
        try:
            sparql = SPARQLWrapper("http://sitaware.isi.edu:8080/bigdata/namespace/wdq/sparql")
            sparql.setQuery(sparql_query)
            sparql.setReturnFormat(JSON)
            results = sparql.query().convert()
            return results['results']['bindings'][0]['x']['value']
        except:
            return ""

    @staticmethod
    def min_distance(word1, word2):
        """
        :type word1: str
        :type word2: str
        :rtype: int
        """
        dp = [[0] * (len(word2) + 1) for _ in range(len(word1) + 1)]
        for i in range(len(word2)):
            dp[0][i + 1] = i + 1
        for i in range(len(word1)):
            dp[i + 1][0] = i + 1
        for i in range(len(word1)):
            for j in range(len(word2)):
                if word1[i] == word2[j]:
                    dp[i + 1][j + 1] = dp[i][j]
                else:
                    dp[i + 1][j + 1] = min(dp[i][j], dp[i + 1][j], dp[i][j + 1]) + 1
        return dp[-1][-1]


if __name__ == "__main__":
    test1 = ['5333265', '5333264', '5333267', '5333266', '5333261', '5333260', '5333263', '5333262']
    test2 = ['XX1007441', 'XX1012281', 'XX1018804', 'XX1033111', 'XX1041185', 'XX1041305', 'XX1044878', 'XX1049656', 'XX1069173']
    test3 = ['0', '1', '1212', '1323', '2112', '2113', '212', '2141', '2142', '2144', '2145', '2147', '215']

    print('Testcase1, Truth P932:')
    for idx, res in enumerate(FindIdentity.get_identifier_3(test1)):
        print('Top', idx+1, res[0])
        print(res[1])
    print('-'*30)
    print('Testcase2, Truth P950')
    for idx, res in enumerate(FindIdentity.get_identifier_3(test2)):
        print('Top', idx+1, res[0])
        print(res[1])
    print('-'*30)
    print('Testcase3, Truth P952')
    for idx, res in enumerate(FindIdentity.get_identifier_3(test3)):
        print('Top', idx+1, res[0])
        print(res[1])
    print('-'*30)



from builtins import range
import pandas as pd  # type: ignore
from langdetect import detect
from datamart_isi.profilers.helpers.feature_compute_hih import is_Decimal_Number
import string
import numpy as np  # type: ignore
import re


# till now, this file totally compute 16 types of features

def compute_missing_space(column, feature, feature_list):
    """
    NOTE: this function may change the input column. It will trim all the leading and trailing whitespace.
            if a cell is empty after trimming(which means it only contains whitespaces),
            it will be set as NaN (missing value), and both leading_space and trailing_space will += 1.

    (1). trim and count the leading space and trailing space if applicable.
        note that more than one leading(trailing) spaces in a cell will still be counted as 1.
    (2). compute the number of missing value for a given series (column); store the result into (feature)
    """

    # if one of them is specified, just compute all; since does not increase lot computations
    if (("number_of_values_with_leading_spaces" in feature_list) or
            ("ratio_of_values_with_leading_spaces" in feature_list) or
            ("number_of_values_with_trailing_spaces" in feature_list) or
            ("ratio_of_values_with_trailing_spaces" in feature_list)):

        leading_space = 0
        trailing_space = 0

        for id, cell in column.iteritems():
            if (pd.isnull(cell)):
                continue

            change = False
            trim_leading_cell = re.sub(r"^\s+", "", cell)
            if (trim_leading_cell != cell):
                leading_space += 1
                change = True
            trim_trailing_cell = re.sub(r"\s+$", "", trim_leading_cell)
            if ((trim_trailing_cell != trim_leading_cell) or len(trim_trailing_cell) == 0):
                trailing_space += 1
                change = True

            # change the origin value in data
            if change:
                if (len(trim_trailing_cell) == 0):
                    column[id] = np.nan
                else:
                    column[id] = trim_trailing_cell

        feature["number_of_values_with_leading_spaces"] = leading_space
        feature["ratio_of_values_with_leading_spaces"] = leading_space / column.size
        feature["number_of_values_with_trailing_spaces"] = trailing_space
        feature["ratio_of_values_with_trailing_spaces"] = trailing_space / column.size


def compute_length_distinct(column, feature, delimiter, feature_list):
    """
    two tasks because of some overlaping computation:

    (1). compute the mean and std of length for each cell, in a given series (column);
        mean and std precision: 5 after point
        missing value (NaN): treated as does not exist
    (2). also compute the distinct value and token:
        number: number of distinct value or tokens, ignore the NaN
        ratio: number/num_total, ignore all NaN
    """
    # (1)
    column = column.dropna()  # get rid of all missing value
    if (column.size == 0):  # if the column is empty, do nothing
        return

    # 1. for character
    # lenth_for_all =  column.apply(len)
    # feature["string_length_mean"] = lenth_for_all.mean()
    # feature["string_length_std"] = lenth_for_all.std()

    # (2)
    if (("number_of_distinct_values" in feature_list) or
            ("ratio_of_distinct_values" in feature_list)):
        feature["number_of_distinct_values"] = column.nunique()
        feature["ratio_of_distinct_values"] = feature["number_of_distinct_values"] / float(column.size)

    if (("number_of_distinct_tokens" in feature_list) or
            ("ratio_of_distinct_tokens" in feature_list)):
        tokenlized = pd.Series([token for lst in column.str.split().dropna() for token in lst])  # tokenlized Series
        lenth_for_token = tokenlized.apply(len)
        # feature["token_count_mean"] = lenth_for_token.mean()
        # feature["token_count_std"] = lenth_for_token.std()
        feature["number_of_distinct_tokens"] = tokenlized.nunique()
        feature["ratio_of_distinct_tokens"] = feature["number_of_distinct_tokens"] / float(tokenlized.size)


def compute_lang(column, feature):
    """
    compute which language(s) it use for a given series (column); store the result into (feature).
    not apply for numbers

    PROBLEMS:
    1. not accurate when string contains digits
    2. not accurate when string is too short
    maybe need to consider the special cases for the above conditions
    """
    column = column.dropna()  # ignore all missing value
    if (column.size == 0):  # if the column is empty, do nothing
        return

    feature["natural_language_of_feature"] = list()
    language_count = {}

    for cell in column:
        if cell.isdigit() or is_Decimal_Number(cell):
            continue
        else:
            # detecting language
            try:
                language = detect(cell)
                if language in language_count:
                    language_count[language] += 1
                else:
                    language_count[language] = 1
            except Exception as e:
                print("there is something may not be any language nor number: {}".format(cell))
                pass

    languages_ordered = sorted(language_count, key=language_count.get, reverse=True)
    for lang in languages_ordered:
        lang_obj = {}
        lang_obj['name'] = lang
        lang_obj['count'] = language_count[lang]
        feature["natural_language_of_feature"].append(lang_obj)


def compute_filename(column, feature):
    """
    compute number of cell whose content might be a filename
    """
    column = column.dropna()  # ignore all missing value

    filename_pattern = r"^\w+\.[a-z]{1,5}"
    column.str.match(filename_pattern)
    num_filename = column.str.match(filename_pattern).sum()
    feature["num_filename"] = num_filename


def compute_punctuation(column, feature, weight_outlier):
    """
    compute the statistical values related to punctuations, for details, see the format section of README.

    not apply for numbers (eg: for number 1.23, "." does not count as a punctuation)

    weight_outlier: = number_of_sigma in function "helper_outlier_calcu"
    """

    column = column.dropna()  # get rid of all missing value
    if (column.size == 0):  # if the column is empty, do nothing
        return

    number_of_chars = sum(column.apply(len))  # number of all chars in column
    num_chars_cell = np.zeros(column.size)  # number of chars for each cell
    puncs_cell = np.zeros([column.size, len(string.punctuation)],
                          dtype=int)  # (number_of_cell * number_of_puncs) sized array

    # step 1: pre-calculations
    cell_id = -1
    for cell in column:
        cell_id += 1
        num_chars_cell[cell_id] = len(cell)
        # only counts puncs for non-number cell
        if cell.isdigit() or is_Decimal_Number(cell):
            continue
        else:
            counts_cell_punc = np.asarray(list(cell.count(c) for c in string.punctuation))
            puncs_cell[cell_id] = counts_cell_punc

    counts_column_punc = puncs_cell.sum(axis=0)  # number of possible puncs in this column
    cell_density_array = puncs_cell / num_chars_cell.reshape([column.size, 1])
    puncs_density_average = cell_density_array.sum(axis=0) / column.size

    # step 2: extract from pre-calculated data
    # only create this feature when punctuations exist
    if (sum(counts_column_punc) > 0):
        feature["most_common_punctuations"] = list()  # list of dict

        # extract the counts to feature, for each punctuation
        for i in range(len(string.punctuation)):
            if (counts_column_punc[i] == 0):  # if no this punctuation occur in the whole column, ignore
                continue
            else:
                punc_obj = {}
                punc_obj["punctuation"] = string.punctuation[i]
                punc_obj["count"] = counts_column_punc[i]
                punc_obj["ratio"] = counts_column_punc[i] / float(number_of_chars)
                punc_obj["punctuation_density_aggregate"] = {"mean": puncs_density_average[i]}
                # calculate outlier
                outlier_array = helper_outlier_calcu(cell_density_array[:, i], weight_outlier)
                # only one element in outlier
                punc_obj["punctuation_density_outliers"] = [{"n": weight_outlier,
                                                             "count": sum(outlier_array)}]
                feature["most_common_punctuations"].append(punc_obj)

    # step 3: sort
    feature["most_common_punctuations"] = sorted(feature["most_common_punctuations"], key=lambda k: k['count'],
                                                 reverse=True)


def helper_outlier_calcu(array, number_of_sigma):
    """
    input: array is a 1D numpy array, number_of_sigma is a integer.
    output: boolean array, size same with input array; true -> is outlier, false -> not outlier
    outlier def:
        the values that not within mean +- (number_of_sigma * sigma) of the statics of the whole list
    """
    mean = np.mean(array)
    std = np.std(array)
    upper_bound = mean + number_of_sigma * std
    lower_bound = mean - number_of_sigma * std
    outlier = (array > upper_bound) + (array < lower_bound)
    return outlier
